"""


This is a vital component of the Ratio1 ecosystem (Ratio1 Edge Protocol). It is a network monitor plugin
that is responsible for monitoring the network status of the nodes in the network as well as persisting
the network status to the database and triggering the epoch serialization


"""
from naeural_core.business.base import BasePluginExecutor
from naeural_core.business.mixins_admin.network_monitor_mixin import _NetworkMonitorMixin, NMonConst
from naeural_core.constants import NET_MON_01_SUPERVISOR_LOG_TIME

__VER__ = '1.0.1'

_CONFIG = {
  **BasePluginExecutor.CONFIG,

  'ALLOW_EMPTY_INPUTS'            : True,
  
  "PROCESS_DELAY"                 : 10,
  
  "SUPERVISOR"                    : False, # obsolete as of 2025-02-03
  "SUPERVISOR_ALERT_TIME"         : 30,
  "SUPERVISOR_LOG_TIME"           : NET_MON_01_SUPERVISOR_LOG_TIME,
  "SEND_IF_NOT_SUPERVISOR"        : False,
  
  "ALERT_RAISE_CONFIRMATION_TIME" : 1,
  "ALERT_LOWER_CONFIRMATION_TIME" : 2,
  "ALERT_DATA_COUNT"              : 2,
  "ALERT_RAISE_VALUE"             : 0.75,
  "ALERT_LOWER_VALUE"             : 0.4,
  "ALERT_MODE"                    : 'mean',
  "ALERT_MODE_LOWER"              : 'max',
  
  "DEBUG_EPOCHS"                  : False,
  
  
  # debug stuff
  "LOG_INFO"            : False,
  "LOG_FULL_INFO"       : False,
  "EXCLUDE_SELF"        : False,
  #
  # The SAVE_NMON_EACH is used to save the network status to the database and should trigger
  # enough so that the network status "often enough" meaning if there are 5-6 restarts per day
  # each summing up 1 minute (total 5-6 minutes) is way below 1% of the day. 
  # The default value is 12 which means that the network status is saved every 2 minutes
  "SAVE_NMON_EACH"      : 12, # this saves each SAVE_NMON_EACH * PROCESS_DELAY seconds

  'VALIDATION_RULES': {
    **BasePluginExecutor.CONFIG['VALIDATION_RULES'],
  },
}


class NetMon01Plugin(
  BasePluginExecutor,
  _NetworkMonitorMixin,
  ):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    super(NetMon01Plugin, self).__init__(**kwargs)
    self._nmon_counter = 0
    self.__supervisor_log_time = 0
    self.__state_loaded = False
    self.__last_epoch_debug_time = 0
    return

  def startup(self):
    super().startup()
    return

  def _is_truthy_env_value(self, value, default=False):
    """
    Normalize NETMON environment toggles to a boolean value.

    Parameters
    ----------
    value : object
      Raw environment value read from ``os.environ``.
    default : bool, optional
      Value returned when the environment variable is unset.

    Returns
    -------
    bool
      ``True`` only for the accepted truthy textual values used by the admin
      plugins. Unset values fall back to ``default``.
    """
    if value is None:
      return default
    return str(value).lower() in ['true', '1', 'yes']
  
  def on_init(self):
    # the following code was left here for historical reasons only
    # convert supervisor to bool if needed
    is_supervisor = self.cfg_supervisor
    if isinstance(is_supervisor, str):
      self.P("Found string value for SUPERVISOR: {}. Converting to bool".format(is_supervisor))
      self.config_data['SUPERVISOR'] = is_supervisor.lower() == 'true'
    #endif is string
    
    if self.is_supervisor_node != self.cfg_supervisor:
      self.P("Warning: Detected admin pipeline config SUPERVISOR={} on node with is_supervisor_node={}".format(
        self.cfg_supervisor, self.is_supervisor_node), color='r'
      )
      self.config_data['SUPERVISOR'] = self.is_supervisor_node
      self.P("Running with SUPERVISOR={}".format(self.cfg_supervisor))
    
    
    # NOTE: SET THE ENVIRONMENT VARIABLES in Dockerfile for Ratio1 implementations
    
    # variable to check if only online nodes should be sent in CURRENT_NETWORK
    
    _send_only_online = self.os_environ.get(self.const.EE_NETMON_SEND_ONLY_ONLINE_ENV_KEY, False)
    self.__send_only_online = str(_send_only_online).lower() in ['true', '1', 'yes']
    
    # index CURRENT_NETWORK by address or by eeid
    _address_as_index = self.os_environ.get(self.const.EE_NETMON_ADDRESS_INDEX_ENV_KEY, False)
    self.__address_as_index = str(_address_as_index).lower() in ['true', '1', 'yes']
    
    # send current network each random seconds (0 means no random delay and execute each send 
    # of CURRENT_NETWORK at each PROCESS_DELAY)
    # R1: EE_NETMON_SEND_CURRENT_NETWORK_EACH  = 50-70
    # other: 0
    try:
      _send_current_network_each = int(self.os_environ.get(
        self.const.EE_NETMON_SEND_CURRENT_NETWORK_EACH_ENV_KEY, 0
      ))
      _send_current_network_each = (
        _send_current_network_each // 2 + self.np.random.randint(1, _send_current_network_each // 2)
      )
    except:
      _send_current_network_each = 0
    self.__send_current_network_each = _send_current_network_each
    _compress_netmon = self.os_environ.get(self.const.EE_NETMON_COMPRESS_ENV_KEY, None)
    self.__compress_netmon = self._is_truthy_env_value(_compress_netmon, default=self.const.ETH_ENABLED)
    self.__last_current_network_time = 0
    msg = f'Netmon initialised:'
    msg += f'\n  - {self.send_current_network_each=}'
    msg += f'\n  - {self.send_only_online=}'
    msg += f'\n  - {self.compress_netmon=}'
    msg += f'\n  - {self.cfg_supervisor=}'
    msg += f'\n  - {self.is_supervisor_node=}'
    msg += f'\n  - {self.cfg_supervisor_log_time=}'
    self.P(msg)
    return
  
  @property
  def address_as_index(self):
    return self.__address_as_index
  
  @property
  def send_only_online(self):
    return self.__send_only_online
  
  @property 
  def send_current_network_each(self):
    return self.__send_current_network_each

  @property
  def compress_netmon(self):
    return self.__compress_netmon

  def _maybe_prepare_netmon_payload(self, payload):
    """
    Wrap payload serialization so NET_MON compression happens at ``to_dict`` time.

    Parameters
    ----------
    payload : GeneralPayload or None
      Freshly-created NET_MON payload object returned by ``_create_payload``.

    Returns
    -------
    GeneralPayload or None
      The same payload instance when compression is disabled, or the same
      instance with ``to_dict`` replaced by a cached encoder that emits NETMON
      v2 wire dictionaries.

    Notes
    -----
    The wrapper preserves the normal payload creation path and applies the
    compression boundary only after the payload has been materialized through
    ``to_dict``. The encoded dictionary is cached so repeated serialization in
    the same send cycle does not recompress the business body.
    """
    if payload is None or not self.compress_netmon:
      return payload

    original_to_dict = payload.to_dict
    cached_payload = {}

    def compressed_to_dict():
      if "payload" not in cached_payload:
        payload_dict = original_to_dict()
        cached_payload["payload"] = self.const.PAYLOAD_DATA.maybe_encode_netmon_payload(
          payload_dict,
          log=self.log,
        )
      return self.deepcopy(cached_payload["payload"])

    payload.to_dict = compressed_to_dict
    return payload
  

  def _maybe_load_state(self):
    """
    This is partially redundant due to the fact that netmon already loads state
    in its bootup. However, this is a safety measure to ensure that the state is loaded.
    """
    if self.netmon.state_already_loaded:
      return
    self.netmon.network_load_status()
    return
  
  def on_command(self, data, **kwargs):

    request_type = 'history' #default to history
    target_node = None
    target_addr = None
    request_type = 'history'
    request_options = {}
    if isinstance(data, dict):
      dct_cmd = {k.lower() : v for k,v in data.items()} # lower case instance command keys
      target_node = dct_cmd.get('node', None)
      target_addr = dct_cmd.get('addr', None)
      request_type = dct_cmd.get('request', 'history')
      request_options = dct_cmd.get('options', {})

    if target_node is not None:
      target_addr = target_addr or self.netmon.network_node_addr(target_node)

    if target_addr is not None:
      self.P("Network monitor on {} ({}) received request for {} ({}): {}".format(
        self.e2_addr, self.eeid, target_addr, target_node, data))
      self._exec_netmon_request(
        target_addr=target_addr,
        request_type=request_type,
        request_options=request_options,
        data=data,
      )
    else:
      self.P("Network monitor on {} ({}) received invalid request for {} ({}): {}".format(
        self.e2_addr, self.eeid, target_addr, target_node, data), color='r')
    return
  
  def _maybe_save_debug_epoch(self):
    if self.cfg_debug_epochs and self.time() - self.__last_epoch_debug_time > 3600: # 1 hour
      self.__last_epoch_debug_time = self.time()
      epoch_manager = self.netmon.epoch_manager
      epoch_node_list = epoch_manager.get_node_list()
      epoch_node_states = [epoch_manager.get_node_state(node) for node in epoch_node_list]
      epoch_node_states = self.deepcopy(epoch_node_states)
      for entry in epoch_node_states:
        entry['current_epoch']['hb_dates'] = sorted(entry['current_epoch']['hb_dates'])
      epoch_node_epochs = [epoch_manager.get_node_epochs(node) for node in epoch_node_list]
      epoch_node_previous_epoch = [epoch_manager.get_node_previous_epoch(node) for node in epoch_node_list]
      epoch_node_last_epoch = [epoch_manager.get_node_last_epoch(node) for node in epoch_node_list]
      epoch_node_first_epoch = [epoch_manager.get_node_first_epoch(node) for node in epoch_node_list]
      epoch_stats = epoch_manager.get_stats()
      debug_epoch={
        "node_list": epoch_node_list,
        "node_states": epoch_node_states,
        "node_epochs": epoch_node_epochs,
        "node_previous_epoch": epoch_node_previous_epoch,
        "node_last_epoch": epoch_node_last_epoch,
        "node_first_epoch": epoch_node_first_epoch,
        "stats": epoch_stats,
      }
      self.log.save_output_json(
        data_json=debug_epoch,
        fname="{}.json".format(self.now_str(short=True)),
        subfolder_path="debug_epoch",
        indent=True,
      )
    return

  
  def process_whitelists(self, current_network):
    """
    this function constructs a full whitelist from all nodes in the current network then
    replaces the individual whitelists with indexes to the full whitelist
    the full whitelist is sorted so that indexes are consistent across nodes
    """
    dct_whitelist = {}
    if self.const.ETH_ENABLED:
      full_whitelist = set()
      for addr in current_network:
        whitelist = current_network[addr].get(self.const.PAYLOAD_DATA.NETMON_WHITELIST, [])
        full_whitelist.update(whitelist)
      lst_whitelist = sorted(list(full_whitelist))
      dct_whitelist = {addr:idx for idx,addr in enumerate(lst_whitelist)}
      for addr in current_network:
        whitelist = current_network[addr].get(self.const.PAYLOAD_DATA.NETMON_WHITELIST, [])
        idxs = [dct_whitelist[w] for w in whitelist if w in dct_whitelist]
        current_network[addr][self.const.PAYLOAD_DATA.NETMON_WHITELIST] = idxs
    return current_network, dct_whitelist
  

  def _process(self):
    payload = None
    self.netmon.epoch_manager.maybe_update_cached_data()
    self._nmon_counter += 1      
    self._maybe_load_state()
    
    # TODO: change to addresses later
    current_nodes, new_nodes = self._add_to_history()       
    ranking = self._get_rankings()    
    
    str_ranking = ", ".join(["{}:{:.0f}:{:.1f}s".format(a,b,c) for a,b,c in ranking])
    
    
    is_anomaly, alerted_nodes = False, None
    
    current_network = None
    current_alerted = None
    is_supervisor = False
    current_ranking = ranking
    current_new = new_nodes
    if not self.cfg_supervisor:
      if (self._nmon_counter % 30) == 0: 
        self.P("Saving local epoch manager status ...")
        self.netmon.epoch_manager.save_status()
      #endif save status if not supervisor
    else: # supervisor
      # save status
      save_nmon_each = int(min(self.cfg_save_nmon_each, 20)) # default no more than 20 iterations
      if (self._nmon_counter % save_nmon_each) == 0: 
        self.P("Saving netmon status for {} nodes".format(len(current_nodes)))
        self.netmon.network_save_status()
      #endif save status
      self.netmon.epoch_manager._maybe_calculate_stats()

      is_anomaly, alerted_nodes = self._supervisor_check()
      self.alerter_add_observation(int(is_anomaly))        
      current_network = current_nodes
      current_alerted = alerted_nodes
      is_supervisor = True  

      configured_supervisor_log_time = self.cfg_supervisor_log_time
      elapsed = self.time() - self.__supervisor_log_time
      if configured_supervisor_log_time is not None and elapsed > configured_supervisor_log_time:
        str_log = "***** Supervisor node sending status for network of {} nodes *****".format(len(current_network))
        known_nodes = self.netmon.network_known_nodes()
        for addr in known_nodes:
          eeid = self.netmon.network_node_eeid(addr)
          str_eeid = "'{}'".format(eeid[:10])
          str_eeid = "{:<13}".format(str_eeid)
          node_info = known_nodes[addr]
          _key = addr if self.address_as_index else eeid
          working_status = current_network.get(_key, {}).get(
            self.const.PAYLOAD_DATA.NETMON_STATUS_KEY, False
          )
          pipelines = node_info['pipelines']
          last_received = node_info['timestamp']
          ago = "{:5.1f}".format(round(self.time() - last_received, 2))
          ago = ago.strip()[:5]
          str_log += "\n - Node: <{}> {} ago {}s had {} pipelines, status: {}".format(
            addr, str_eeid, ago, len(pipelines), working_status
          )
        nr_dir, nr_hb = self.netmon.get_hb_vs_direct_pipeline_sources()
        str_log +=   "\n   Sources: {} direct, {} hb".format(nr_dir, nr_hb)
        self.P(str_log)
        self.__supervisor_log_time = self.time()
      #endif supervisor log time       
    #endif supervisor or not
    
    self._maybe_save_debug_epoch()

    should_send = (self.time() - self.__last_current_network_time) > self.send_current_network_each
    if (self.cfg_supervisor or self.cfg_send_if_not_supervisor) and should_send:
      if self.send_only_online:
        # we only send online nodes or nodes are allowed by this node (thus might be oracles)
        current_network = {
          k : v for k,v in current_network.items() 
          if (
            v.get(self.const.PAYLOAD_DATA.NETMON_STATUS_KEY, "") == self.const.PAYLOAD_DATA.NETMON_STATUS_ONLINE 
            or
            self.bc.is_node_allowed(v.get(self.const.PAYLOAD_DATA.NETMON_ADDRESS))
          )
        }    
      message="" if len(current_alerted) == 0 else "Missing/lost processing nodes: {}".format(list(current_alerted.keys()))
      # compress whitelists if needed  
      n_nodes = len(current_network)
      netsize = len(self.json_dumps(current_network))
      if self.const.ETH_ENABLED:
        # this is a EVM-based implementation so we need to process whitelists
        current_network, dct_whitelist = self.process_whitelists(current_network)
        n_nodes = len(current_network)
        netsize2 = len(self.json_dumps(current_network))
        self.P("Compressed wl for {} nodes from {} to {} bytes (reduction {:.1f}%)".format(
          n_nodes, netsize, netsize2, 100.0 * (netsize - netsize2) / netsize if netsize > 0 else 0.0
        ))
      else:
        # no whitelist processing
        self.P("Not processing wl for {} nodes (total {} bytes full data, ETH_ENABLED={})...".format(
          n_nodes, netsize, self.const.ETH_ENABLED
        ))        
        dct_whitelist = {}
      #endif eth enabled
      # for this plugin only ALERTS should be used in UI/BE
      payload = self._create_payload(
        current_server=self.e2_addr,
        current_network=current_network,
        current_alerted=current_alerted,
        whitelist_map=dct_whitelist,
        message=message,
        status=message,
        current_ranking=current_ranking,
        current_new=current_new,
        is_supervisor=is_supervisor,
        send_current_network_each=self.send_current_network_each,
      )
      payload = self._maybe_prepare_netmon_payload(payload)
      self.__last_current_network_time = self.time()
    #endif should send 

    if self.cfg_log_full_info:
      self.P("Full info:\n{}".format(self.json.dumps(current_nodes, indent=4)))

    if self.cfg_log_info:
      self.P("Anomaly: {}, IsAlert: {}, IsNewAlert: {}, Alerted: {}, Ranking: {}".format(
        is_anomaly, self.alerter_is_alert(), self.alerter_is_new_alert(),
        list(alerted_nodes.keys()), str_ranking
      ))
      self.P("Alerter status: {}".format(self.get_alerter_status()))
    if self.alerter_is_new_alert():
      self.P("NetMon anomaly:\n********************\nAnomaly reported for {} nodes:\n{}\n ********************".format(
        len(alerted_nodes), self.json_dumps(alerted_nodes, indent=2)
      ))
    #endif show alerts
    return payload



"""
chainstore_response_mixin.py

A reusable mixin that provides chainstore response functionality for plugins.

This mixin implements a standard pattern for sending plugin startup confirmations
or other lifecycle events to the chainstore, enabling asynchronous callback mechanisms
for distributed plugin orchestration.

Design Pattern:
--------------
This follows the Mixin Pattern, which allows plugins to compose behaviors by
inheriting from multiple specialized classes. The mixin provides:

1. Template Method Pattern: reset/set methods as templates
2. Strategy Pattern: Subclasses can override _get_chainstore_response_data()
3. Observer Pattern: Chainstore acts as the message broker for observers

Usage (Automatic - via BasePluginExecutor):
-------------------------------------------
This mixin is automatically included in BasePluginExecutor. For simple plugins,
no action is required - chainstore response is sent automatically after on_init().

For complex plugins that need deferred readiness (containers, APIs, etc.),
use set_plugin_ready() from _PluginReadinessMixin:
```python
class MyComplexPlugin(BasePlugin):
  def on_init(self):
    super().on_init()
    self.set_plugin_ready(False)  # Defer until truly ready
    # ... start async initialization ...

  def _after_service_ready(self):
    self.set_plugin_ready(True)  # Triggers both chainstore AND semaphore
```

For custom response data:
```python
class MyPlugin(BasePlugin):
  def _get_chainstore_response_data(self):
    data = super()._get_chainstore_response_data()
    data.update({
      'api_port': self.cfg_port,
      'custom_field': self.custom_value,
    })
    return data
```

Architecture Benefits:
---------------------
1. Single Responsibility: Mixin only handles chainstore response logic
2. Open/Closed: Plugins can extend response data without modifying mixin
3. DRY: Eliminates code duplication across multiple plugin types
4. Testability: Mixin can be tested independently
5. Composability: Can be mixed with other functionality mixins
6. Simplicity: Single write - no retries, no confirmations
7. Process Loop Pattern: Like semaphore, checks readiness in _process()

Configuration:
-------------
CHAINSTORE_RESPONSE_KEY (str, optional):
  The key under which to store the response in chainstore.
  If None or not set, no response will be sent/reset.
  This is typically set by orchestration systems like Deeploy.

Security Considerations:
-----------------------
- Response keys should be generated with sufficient entropy to prevent guessing
- Response data should not contain sensitive information (passwords, tokens, etc.)

"""


class _ChainstoreResponseMixin:
  """
  Mixin providing chainstore response functionality for plugin lifecycle events.

  This mixin enables plugins to send confirmation data to a distributed chainstore
  when important lifecycle events occur (e.g., plugin startup, state changes).

  The mixin uses the Template Method pattern to provide a standard flow while
  allowing subclasses to customize the response data through hook methods.

  Key principle: Reset at start, send when ready (via process loop).
  """

  def __init__(self):
    """
    Initialize chainstore response state variables.

    State variables:
      - _chainstore_response_sent: Prevents duplicate sends

    Note: _is_plugin_ready and set_plugin_ready() are now in _PluginReadinessMixin
    """
    self._chainstore_response_sent = False
    super(_ChainstoreResponseMixin, self).__init__()
    return


  def _chainstore_maybe_auto_send(self):
    """
    Automatically send chainstore response when plugin is ready.

    Called from process loop (_process). Sends only once.
    This follows the same pattern as _semaphore_maybe_auto_signal().

    Uses is_plugin_ready() from _PluginReadinessMixin which resolves:
      - None: use default (uvicorn_server_started or _init_process_finalized)
      - False: explicitly deferred, wait for set_plugin_ready(True)
      - True: explicitly ready
    """
    # No key configured
    if not self._should_send_chainstore_response():
      return

    # Already sent
    if self._chainstore_response_sent:
      return

    # Check unified readiness (from _PluginReadinessMixin)
    if not self.is_plugin_ready():
      return

    # Ready - send response
    if self._send_chainstore_response():
      self._chainstore_response_sent = True
    return

  def _get_chainstore_response_key(self):
    """
    Get the chainstore response key from configuration.

    This method follows the Dependency Inversion Principle by depending on
    configuration abstraction rather than concrete implementation details.

    Returns:
        str or None: The response key if configured, None otherwise.
    """
    return getattr(self, 'cfg_chainstore_response_key', None)

  def _get_chainstore_response_data(self):
    """
    Template method hook: Build the response data dictionary.

    This method can be overridden by subclasses to provide custom response data.
    The default implementation returns base plugin information.

    Design Pattern: Template Method Pattern
    - This is the "hook" method that subclasses can override
    - The parent method _send_chainstore_response() is the "template"

    Best Practice: When overriding, call super() first then extend:
    ```python
    def _get_chainstore_response_data(self):
      data = super()._get_chainstore_response_data()
      data.update({
        'custom_field': self.custom_value,
      })
      return data
    ```

    Returns:
        dict: Response data to be stored in chainstore.
              Should be JSON-serializable.

    Security Note:
        Never include sensitive data like passwords, private keys, or tokens
        in the response data. This data may be visible to multiple nodes.
    """
    # Basic data always available
    data = {
      'plugin_signature': self.get_signature(),
      'instance_id': self.get_instance_id(),
      'timestamp': self.time_to_str(self.time()),
    }

    # Add base plugin fields (from BasePluginExecutor)
    data['stream_id'] = self.get_stream_id()
    data['plugin_version'] = self.__version__
    data['node_id'] = self.ee_id
    data['node_addr'] = self.ee_addr

    # Default status (can be overridden by subclasses)
    if 'status' not in data:
      data['status'] = 'ready'
      data['is_ready'] = True

    # Merge optional custom fields without forcing subclasses to override this method
    extra = {}
    try:
      extra = self.get_chainstore_response()
      if extra is None:
        extra = {}
      elif not isinstance(extra, dict):
        self.P(
          f"get_chainstore_response() must return a dict, got {type(extra)}",
          color='r'
        )
        extra = {}
    except Exception as exc:
      self.P(f"Error in get_chainstore_response(): {exc}", color='r')
      extra = {}

    data.update(extra)
    return data


  def get_chainstore_response(self):
    """
    Public hook to add custom fields to chainstore response data.

    Override this method to append additional JSON-serializable fields
    without replacing the standard response keys provided by
    _get_chainstore_response_data().

    Returns
    -------
    dict
        Extra fields to merge into the response (default: {}).
    """
    return {}


  def reset_chainstore_response(self):
    """
    Reset chainstore response state for plugin restart/recycle.

    Call this when preparing for a restart to:
    1. Clear the chainstore key (signals "restarting" to orchestrators)
    2. Allow a new response to be sent when ready again

    Typically paired with set_plugin_ready(False).

    Returns
    -------
    bool
        True if reset was performed, False if key not configured.

    Example
    -------
    ```python
    def _restart_container(self):
        self.reset_chainstore_response()
        self.set_plugin_ready(False)
        # ... perform restart ...
    ```
    """
    return self._reset_chainstore_response()


  def _should_send_chainstore_response(self):
    """
    Determine if a chainstore response should be sent.

    This method implements validation logic to ensure responses are only
    sent when properly configured. Can be overridden for custom logic.

    Returns:
        bool: True if response should be sent, False otherwise.
    """
    response_key = self._get_chainstore_response_key()
    if response_key is None:
      return False

    if not isinstance(response_key, str) or len(response_key) == 0:
      self.P(
        "CHAINSTORE_RESPONSE_KEY is configured but invalid (must be non-empty string)",
        color='r'
      )
      return False

    return True

  def _reset_chainstore_response(self):
    """
    Reset (clear) the chainstore response key at plugin start.

    This should be called at the very beginning of plugin initialization to
    signal that the plugin is starting up. The orchestration system can monitor
    this key - if it's None/empty, it means the plugin is still initializing.

    After successful initialization, call _send_chainstore_response() to set
    the actual response data.

    Returns:
        bool: True if reset was performed, False if key not configured.

    Example:
        ```python
        def on_init(self):
            super().on_init()
            self._reset_chainstore_response()  # Clear at start
            # ... initialization code ...
            self._send_chainstore_response()    # Set after success
            return
        ```
    """
    if not self._should_send_chainstore_response():
      return False

    response_key = self._get_chainstore_response_key()
    self.P(f"Resetting chainstore response key '{response_key}'")

    # Reset the sent flag to allow re-sending after reset
    self._chainstore_response_sent = False

    try:
      # Set to None to signal "initializing" state
      result = self.chainstore_set(response_key, None)
      if result:
        self.P(f"Successfully reset chainstore key '{response_key}'")
        return True
      else:
        self.P(f"Failed to reset chainstore key '{response_key}'", color='y')
        return False
    except Exception as e:
      self.P(f"Error resetting chainstore key '{response_key}': {e}", color='r')
      return False

  def _send_chainstore_response(self):
    """
    Send plugin response data to chainstore (single write).

    This is the main template method that sends the response after successful
    plugin initialization. It should be called exactly once at the end of
    on_init() after all setup is complete.

    Design Pattern: Template Method Pattern
    - Defines the skeleton of the algorithm
    - Delegates data building to hook method (_get_chainstore_response_data)

    Args:
        None

    Returns:
        bool: True if response was sent successfully, False otherwise.

    Example:
        ```python
        # Send default response data
        self._send_chainstore_response()
        ```

    Implementation Notes:
        - Single write (no retries, no confirmations)
        - Gracefully handles chainstore_set failures without raising exceptions
        - Call _reset_chainstore_response() at plugin start before calling this
    """
    # Validation: Check if response should be sent
    if not self._should_send_chainstore_response():
      return False

    response_key = self._get_chainstore_response_key()

    self.P(f"Sending chainstore response to key '{response_key}'", color='b')

    # Build response data using template method hook
    try:
      response_data = self._get_chainstore_response_data()

    except Exception as e:
      self.P(
        f"Error building chainstore response data: {e}",
        color='r'
      )
      return False

    # Send single write to chainstore
    try:
      self.P(f"Setting '{response_key}' to: {self.json_dumps(response_data)}")

      # Single write - no retries, no confirmations
      result = self.chainstore_set(response_key, response_data, debug=True)

      if result:
        self.P(f"Successfully sent chainstore response to '{response_key}'", color='g')
        return True
      else:
        self.P(f"Failed to send chainstore response (chainstore_set returned False)", color='y')
        return False

    except Exception as e:
      self.P(f"Error sending chainstore response: {e}", color='r')
      return False



import ast
import pathlib
import unittest


FASTAPI_WEBAPP_PATH = pathlib.Path(__file__).resolve().parents[1] / "default" / "web_app" / "fast_api_web_app.py"


class TestFastApiWebAppSemaphoreEnv(unittest.TestCase):

  @classmethod
  def setUpClass(cls):
    cls.source = FASTAPI_WEBAPP_PATH.read_text()
    cls.module = ast.parse(cls.source)

  def _get_setup_semaphore_env(self):
    for node in self.module.body:
      if isinstance(node, ast.ClassDef) and node.name == "FastApiWebAppPlugin":
        for item in node.body:
          if isinstance(item, ast.FunctionDef) and item.name == "_setup_semaphore_env":
            return item
    self.fail("FastApiWebAppPlugin._setup_semaphore_env not found")

  def test_fastapi_webapp_defines_default_semaphore_env_hook(self):
    method = self._get_setup_semaphore_env()
    self.assertIsNotNone(method)

  def test_fastapi_webapp_uses_runtime_port_in_semaphore_env_hook(self):
    method = self._get_setup_semaphore_env()
    segment = ast.get_source_segment(self.source, method) or ""
    self.assertIn("port = self.port", segment)
    self.assertNotIn("port = self.cfg_port", segment)

  def test_fastapi_webapp_exports_api_endpoint_keys(self):
    method = self._get_setup_semaphore_env()
    segment = ast.get_source_segment(self.source, method) or ""
    self.assertIn("self.semaphore_set_env('API_IP', localhost_ip)", segment)
    self.assertIn("self.semaphore_set_env('API_PORT', str(port))", segment)
    self.assertIn("self.semaphore_set_env('API_URL', 'http://{}:{}'.format(localhost_ip, port))", segment)
    self.assertNotIn("self.semaphore_set_env('HOST', localhost_ip)", segment)
    self.assertNotIn("self.semaphore_set_env('HOST_IP', localhost_ip)", segment)
    self.assertNotIn("self.semaphore_set_env('PORT', str(port))", segment)
    self.assertNotIn("self.semaphore_set_env('HOST_PORT', str(port))", segment)
    self.assertNotIn("self.semaphore_set_env('URL', 'http://{}:{}'.format(localhost_ip, port))", segment)

  def test_fastapi_webapp_guards_host_resolution_failures(self):
    method = self._get_setup_semaphore_env()
    segment = ast.get_source_segment(self.source, method) or ""
    self.assertIn("try:", segment)
    self.assertIn("localhost_ip = self.log.get_localhost_ip()", segment)
    self.assertIn("except Exception as ex:", segment)
    self.assertIn("Skipping API_IP semaphore export", segment)
    self.assertIn("Skipping API_URL semaphore export", segment)


if __name__ == "__main__":
  unittest.main()



import subprocess

from naeural_core.business.base import BasePluginExecutor
from naeural_core.business.mixins_libs.ngrok_mixin import _NgrokMixinPlugin
from naeural_core.business.mixins_libs.cloudflare_mixin import _CloudflareMixinPlugin


_CONFIG = {
  **BasePluginExecutor.CONFIG,

  "TUNNEL_ENGINE": "cloudflare",  # or "cloudflare"

  "VALIDATION_RULES": {
    **BasePluginExecutor.CONFIG['VALIDATION_RULES'],
  },
}

"""
This class is only made for backward compatibility.
"""

class BaseTunnelEnginePlugin(
  _NgrokMixinPlugin,
  _CloudflareMixinPlugin,
  BasePluginExecutor
):
  """
  Base class for tunnel engine plugins, which can be used to create plugins that
  expose methods as endpoints and tunnel traffic through ngrok or cloudflare.
  """
  CONFIG = _CONFIG

  def use_cloudflare(self):
    """
    Check if the plugin is configured to use Cloudflare as the tunnel engine.
    """
    return self.cfg_tunnel_engine.lower() == "cloudflare"

  @property
  def app_url(self):
    """
    Returns the URL of the application based on the tunnel engine being used.
    """
    if self.use_cloudflare():
      return self.app_url_cloudflare
    return self.app_url_ngrok

  def get_default_tunnel_engine_parameters(self):
    if self.use_cloudflare():
      return self.get_default_tunnel_engine_parameters_cloudflare()
    return self.get_default_tunnel_engine_parameters_ngrok()

  def reset_tunnel_engine(self):
    if self.use_cloudflare():
      return self.reset_tunnel_engine_cloudflare()
    return self.reset_tunnel_engine_ngrok()

  def maybe_init_tunnel_engine(self):
    if self.use_cloudflare():
      return self.maybe_init_tunnel_engine_cloudflare()
    return self.maybe_init_tunnel_engine_ngrok()

  def maybe_start_tunnel_engine(self):
    if self.use_cloudflare():
      return self.maybe_start_tunnel_engine_cloudflare()
    return self.maybe_start_tunnel_engine_ngrok()

  def maybe_stop_tunnel_engine(self):
    if self.use_cloudflare():
      return self.maybe_stop_tunnel_engine_cloudflare()
    return self.maybe_stop_tunnel_engine_ngrok()

  def get_setup_commands(self):
    if self.use_cloudflare():
      return self.get_setup_commands_cloudflare()
    return super(BaseTunnelEnginePlugin, self).get_setup_commands_ngrok()

  def get_start_commands(self):
    if self.use_cloudflare():
      return self.get_start_commands_cloudflare()
    return super(BaseTunnelEnginePlugin, self).get_start_commands_ngrok()

  def check_valid_tunnel_engine_config(self):
    if self.use_cloudflare():
      return self.check_valid_tunnel_engine_config_cloudflare()
    return self.check_valid_tunnel_engine_config_ngrok()

  def on_log_handler(self, text, key=None):
    if self.use_cloudflare():
      return self.on_log_handler_cloudflare(text, key)
    return self.on_log_handler_ngrok(text, key)

  def on_init(self):

    self.dct_logs_reader = {}
    self.dct_err_logs_reader = {}

    super(BaseTunnelEnginePlugin, self).on_init()


  def run_tunnel_command(self, command):
    """
    Run a tunnel command in the background using LogReader like in base web app.
    This is a generic implementation for running tunnel engine commands.
    """
    if not command:
      return None
    
    try:
      self.P(f"Running tunnel command: {command}")
      process = subprocess.Popen(
        command,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        bufsize=0  # this is important for real-time output
      )
      
      logs_reader = self.LogReader(process.stdout, size=100, daemon=None)
      err_logs_reader = self.LogReader(process.stderr, size=100, daemon=None)
      
      # Store the readers for later cleanup
      if not hasattr(self, 'dct_logs_reader'):
        self.dct_logs_reader = {}
      if not hasattr(self, 'dct_err_logs_reader'):
        self.dct_err_logs_reader = {}
      
      self.dct_logs_reader['tunnel'] = logs_reader
      self.dct_err_logs_reader['tunnel'] = err_logs_reader
      
      return process
    except Exception as e:
      self.P(f"Error running tunnel command: {e}")
      return None

  def stop_tunnel_command(self, process):
    """
    Stop a running tunnel command process and clean up LogReaders.
    """
    if process and process.poll() is None:
      try:
        process.terminate()
        process.wait(timeout=5)
      except subprocess.TimeoutExpired:
        process.kill()
        process.wait()
      except Exception as e:
        self.P(f"Error stopping tunnel command: {e}")
    
    # Clean up LogReaders
    self._cleanup_tunnel_log_readers()
    return

  def _cleanup_tunnel_log_readers(self):
    """
    Clean up tunnel LogReaders like in base web app.
    """
    if hasattr(self, 'dct_logs_reader') and 'tunnel' in self.dct_logs_reader:
      logs_reader = self.dct_logs_reader.get('tunnel')
      if logs_reader is not None:
        logs_reader.stop()
        # Read any remaining logs
        logs = logs_reader.get_next_characters()
        if len(logs) > 0:
          self.on_log_handler(logs)
      # end if logs_reader
      self.dct_logs_reader.pop('tunnel', None)

    if hasattr(self, 'dct_err_logs_reader') and 'tunnel' in self.dct_err_logs_reader:
      err_logs_reader = self.dct_err_logs_reader.get('tunnel')
      if err_logs_reader is not None:
        err_logs_reader.stop()
        # Read any remaining error logs
        err_logs = err_logs_reader.get_next_characters()
        if len(err_logs) > 0:
          self.P(f"[stderr][tunnel]: {err_logs}")
      self.dct_err_logs_reader.pop('tunnel', None)
      # end if err_logs_reader
    return

  def read_tunnel_logs(self):
    """
    Read tunnel logs from LogReaders like in base web app.
    """
    if hasattr(self, 'dct_logs_reader') and 'tunnel' in self.dct_logs_reader:
      logs_reader = self.dct_logs_reader.get('tunnel')
      if logs_reader is not None:
        logs = logs_reader.get_next_characters()
        if len(logs) > 0:
          self.on_log_handler(logs)
      # end if logs_reader

    if hasattr(self, 'dct_err_logs_reader') and 'tunnel' in self.dct_err_logs_reader:
      err_logs_reader = self.dct_err_logs_reader.get('tunnel')
      if err_logs_reader is not None:
        err_logs = err_logs_reader.get_next_characters()
        if len(err_logs) > 0:
          self.P(f"[stderr][tunnel]: {err_logs}")
      # end if err_logs_reader
    return


  def run_tunnel_engine(self):
    """
    Run the tunnel engine start command based on the configured engine.
    This is a generic wrapper that delegates to the appropriate tunnel engine.
    """
    if self.use_cloudflare():
      cloudflare_command = self._get_cloudflare_start_command()
      if cloudflare_command:
        return self.run_tunnel_command(cloudflare_command)
    else:
      # For ngrok or other engines
      ngrok_command = self._get_ngrok_start_command()
      if ngrok_command:
        return self.run_tunnel_command(ngrok_command)
    
    return None



"""
Multi-source CV2 video data capture thread.

This implementation extends the single-stream CV2 DCT to operate on multiple
video sources defined through the `SOURCES` configuration key. It keeps the
per-stream behaviour identical to the single-stream version while:

* Handling reconnections asynchronously per stream so that healthy sources keep
  producing data even when others are reconnecting.
* Exposing metadata per captured frame that identifies the originating source.
* Feeding downstream consumers with batched inputs containing one entry per
  active stream.

The implementation mirrors the single-stream DCT as closely as possible so the
overall heuristics (AMD timing, crop/resize, notifications) behave the same for
each individual stream.
"""

from __future__ import annotations

import traceback
from dataclasses import dataclass, field
from threading import Lock, Thread
from typing import Dict, Iterable, List, Optional

import cv2
import numpy as np

from naeural_core import constants as ct
from naeural_core.data.base import DataCaptureThread
from naeural_core.data_structures import MetadataObject


@dataclass
class _StreamContext:
  name: str
  index: int
  config: Dict
  url: str
  metadata: MetadataObject = field(default_factory=MetadataObject)
  capture: Optional[cv2.VideoCapture] = None
  has_connection: bool = False
  reconnecting: bool = False
  last_url: Optional[str] = None
  nr_connection_issues: int = 0
  crt_frame: int = 0
  configured_size_error: bool = False
  resize_hw: Optional[tuple] = None
  frame_crop: Optional[List[int]] = None
  reconnect_thread: Optional[Thread] = field(default=None, repr=False)
  active: bool = True
  lock: Lock = field(default_factory=Lock, repr=False)


def _ensure_positive_randint_bounds(base_val: int) -> int:
  base_val = int(base_val)
  return max(1, base_val)


_CONFIG = {
  **DataCaptureThread.CONFIG,

  "MAX_RETRIES": 2,

  "AMD_TARGET_DPS": 0,
  "SIMULATE_AMD": False,

  "CAP_PROP_BUFFERSIZE": 0,

  "USE_FFMPEG": False,

  "CONFIGURED_H": -1,
  "CONFIGURED_W": -1,

  "SOURCES": [],

  "VALIDATION_RULES": {
    **DataCaptureThread.CONFIG["VALIDATION_RULES"],

    "AMD_TARGET_DPS": {
      "DESCRIPTION": (
        "For AMD architectures forces number of grabs per iteration based on "
        "the target required DPS."
      ),
      "TYPE": "int",
    },
    "SOURCES": {
      "DESCRIPTION": (
        "List of video source descriptors. Each entry must provide at least "
        "the `URL` key and can optionally override other capture parameters."
      ),
      "TYPE": "list",
    },
  },
}


class MultiVideoStreamCv2DataCapture(DataCaptureThread):
  CONFIG = _CONFIG

  def __init__(self, **kwargs):
    self._streams: Dict[str, _StreamContext] = {}
    self._streams_lock = Lock()
    self._metadata = MetadataObject(streams={})
    super(MultiVideoStreamCv2DataCapture, self).__init__(**kwargs)
    return

  def startup(self):
    super().startup()
    self._update_global_metadata()
    new_contexts = self._sync_streams()
    for ctx in new_contexts:
      self._launch_reconnect(ctx)
    return

  def _init(self):
    if not self.has_cap_resolution_config:
      str_err = "Cannot start MultiVideoStreamCv2DataCapture without CAP_RESOLUTION!"
      self.P(str_err, color="error")
      raise ValueError(str_err)
    return

  @property
  def is_hw_error(self):
    return any(ctx.configured_size_error for ctx in self._streams.values())

  def is_intel(self):
    if self.cfg_simulate_amd:
      result = False
    else:
      proc = self.log.get_processor_platform()
      if proc is not None:
        result = "intel" in proc.lower()
      else:
        result = False
    return result

  def _release(self):
    with self._streams_lock:
      for ctx in list(self._streams.values()):
        self._release_stream(ctx)
      self._streams.clear()
    return

  def _release_stream(self, stream: _StreamContext):
    stream.active = False
    with stream.lock:
      if stream.capture is not None:
        try:
          stream.capture.release()
        except Exception:
          pass
        stream.capture = None
      stream.has_connection = False
    stream.reconnecting = False
    stream.reconnect_thread = None
    return

  def _on_config_changed(self):
    new_contexts = self._sync_streams()
    for ctx in new_contexts:
      self._launch_reconnect(ctx)
    return

  def _sync_streams(self) -> List[_StreamContext]:
    sources = self.cfg_sources or []
    if not isinstance(sources, list):
      raise ValueError("SOURCES must be a list")

    normalized: List[tuple] = []
    used_names = set()
    for idx, raw_cfg in enumerate(sources):
      if isinstance(raw_cfg, str):
        cfg = {"URL": raw_cfg}
      elif isinstance(raw_cfg, dict):
        cfg = raw_cfg.copy()
      else:
        self.P(f"Ignoring invalid source definition at index {idx}: {raw_cfg}", color="r")
        continue

      url = cfg.get("URL")
      if not url:
        self.P(f"Ignoring source at index {idx} due to missing URL", color="r")
        continue

      proposed_name = cfg.get("NAME") or f"{self.cfg_name}_src_{idx}"
      name = proposed_name
      suffix = 1
      while name in used_names:
        name = f"{proposed_name}_{suffix}"
        suffix += 1
      used_names.add(name)
      normalized.append((name, idx, cfg, url))

    new_contexts: List[_StreamContext] = []
    with self._streams_lock:
      existing_names = set(self._streams.keys())
      normalized_names = {name for name, _, _, _ in normalized}

      removed_names = existing_names - normalized_names
      for name in removed_names:
        ctx = self._streams.pop(name)
        self.P(f"Removing video source '{name}'", color="y")
        self._release_stream(ctx)

      for name, idx, cfg, url in normalized:
        if name not in self._streams:
          ctx = _StreamContext(name=name, index=idx, config=cfg, url=url)
          ctx.metadata.update(stream_name=name, stream_index=idx, url=url)
          self._streams[name] = ctx
          new_contexts.append(ctx)
          self.P(f"Registered new video source '{name}' -> {url}", color="y")
        else:
          ctx = self._streams[name]
          ctx.index = idx
          ctx.config = cfg
          ctx.url = url
          ctx.metadata.update(stream_index=idx, url=url)
          if ctx.last_url is not None and ctx.last_url != url:
            self.P(
              f"[{name}] URL change detected from {ctx.last_url} to {url}, scheduling reconnect...",
              color="y",
            )
            self._mark_stream_disconnected(ctx, force_reconnect=True)

    self._update_global_metadata()
    return new_contexts

  def _maybe_reconnect(self):
    contexts = self._get_streams_snapshot()
    for ctx in contexts:
      if not ctx.active:
        continue

      if ctx.has_connection:
        with ctx.lock:
          capture = ctx.capture
          opened = capture is not None and capture.isOpened()
        if not opened:
          self.P(f"[{ctx.name}] Capture seems closed. Scheduling reconnect.", color="r")
          self._mark_stream_disconnected(ctx, force_reconnect=True)
        continue

      if not ctx.reconnecting:
        self._launch_reconnect(ctx)

    self._update_global_connection_flags()
    return

  def _launch_reconnect(self, stream: _StreamContext):
    if not stream.active:
      return
    if stream.reconnecting:
      return
    if self._stop:
      return

    stream.reconnecting = True
    thread_name = f"{ct.THREADS_PREFIX}reconn_{self.sanitize_name(stream.name)}"
    worker = Thread(target=self._reconnect_stream_worker, args=(stream,), name=thread_name, daemon=True)
    stream.reconnect_thread = worker
    worker.start()
    return

  def _reconnect_stream_worker(self, stream: _StreamContext):
    round_info = None
    while not self._stop and stream.active and not stream.has_connection:
      max_retries = self._get_stream_cfg_value(stream, "MAX_RETRIES", self.cfg_max_retries)
      if stream.nr_connection_issues == 0:
        msg = f"[{stream.name}] Connecting to url: {stream.url} (deq size: {self._deque.maxlen})"
        color = "g"
      else:
        msg = f"[{stream.name}] Reconnecting ({stream.nr_connection_issues}) to url: {stream.url}"
        color = "r"
      self.P(msg, color=color)

      nr_retry = 0
      round_info = None
      while nr_retry <= max_retries and not self._stop and stream.active and not stream.has_connection:
        self.sleep(1)
        nr_retry += 1
        stream.nr_connection_issues += 1
        try:
          self._release_stream_capture(stream)
          capture = self._open_stream(stream)
          if capture is None or not capture.isOpened():
            self.P(f"[{stream.name}]    Capture failed.", color="r")
            continue

          has_frame, frame = capture.read()
          if not has_frame or frame is None:
            self.P(f"[{stream.name}]    Capture read failed.", color="r")
            capture.release()
            continue

          skip_frames = self._get_stream_cfg_value(stream, "NR_SKIP_FRAMES", self.cfg_nr_skip_frames)
          if skip_frames:
            capture.set(cv2.CAP_PROP_POS_FRAMES, skip_frames)

          try:
            raw_fps = int(capture.get(cv2.CAP_PROP_FPS))
          except Exception:
            raw_fps = 20
          fps = int(np.clip(raw_fps, 1, 120))
          buff_size = capture.get(cv2.CAP_PROP_BUFFERSIZE)
          height = int(capture.get(cv2.CAP_PROP_FRAME_HEIGHT))
          width = int(capture.get(cv2.CAP_PROP_FRAME_WIDTH))
          frame_count = int(max(0, capture.get(cv2.CAP_PROP_FRAME_COUNT)))

          with stream.lock:
            stream.capture = capture
            stream.has_connection = True
            stream.last_url = stream.url
            stream.crt_frame = 0
            stream.configured_size_error = False
            stream.resize_hw = None
            stream.frame_crop = None
            stream.metadata.update(
              fps=fps,
              frame_h=height,
              frame_w=width,
              frame_count=frame_count,
              buffersize=buff_size,
              stream_name=stream.name,
              stream_index=stream.index,
              url=stream.url,
            )

          self._configure_stream_resize_crop(stream, height, width)
          self.reset_received_first_data()
          self.P(f"[{stream.name}] Connection seems to be established.")
          self._create_stream_notification(stream, success=True, info=None)
          self._update_global_connection_flags()
          self._update_global_metadata()
          stream.reconnecting = False
          return
        except Exception:
          round_info = traceback.format_exc()
          self.P(f"[{stream.name}] `_maybe_reconnect` exception: {round_info}", color="r")
      # end while retries

      sleep_base = _ensure_positive_randint_bounds(self.cfg_sleep_time_on_error)
      sleep_high = max(sleep_base + 1, sleep_base * 3 + 1)
      sleep_time = int(np.random.randint(sleep_base, sleep_high))

      self._create_stream_notification(
        stream,
        success=False,
        info=round_info,
        extra={
          "attempts": nr_retry,
          "sleep_time": sleep_time,
        },
      )

      if not self.is_reconnectable or not stream.active:
        break

      self.sleep(sleep_time)

    stream.reconnecting = False
    self._update_global_connection_flags()
    return

  def _create_stream_notification(self, stream: _StreamContext, success: bool, info=None, extra=None):
    extra = extra or {}
    metadata_snapshot = self.deepcopy(stream.metadata.__dict__)
    metadata_snapshot["has_connection"] = stream.has_connection
    metadata_snapshot["nr_connection_issues"] = stream.nr_connection_issues

    if success:
      msg = (
        f"Video DCT '{self.cfg_name}' successfully connected stream '{stream.name}'. "
        f"Overall {stream.nr_connection_issues} reconnects."
      )
      notification_type = ct.STATUS_TYPE.STATUS_NORMAL
      color = "g"
    else:
      attempts = extra.get("attempts", 0)
      sleep_time = extra.get("sleep_time", 0)
      msg = (
        f"Abnormal functioning of Video DCT '{self.cfg_name}' stream '{stream.name}' "
        f"failed after {attempts} retries (overall {stream.nr_connection_issues} reconnects). "
        f"Sleeping stream reconnect for {sleep_time:.1f}s"
      )
      notification_type = ct.STATUS_TYPE.STATUS_ABNORMAL_FUNCTIONING
      color = "e"

    self.P(msg, color=color)
    self._create_notification(
      notif=notification_type,
      msg=msg,
      info=info,
      video_stream_info=metadata_snapshot,
      stream_name=stream.name,
      displayed=True,
    )
    return

  def _open_stream(self, stream: _StreamContext):
    use_ffmpeg = bool(self._get_stream_cfg_value(stream, "USE_FFMPEG", self.cfg_use_ffmpeg))
    if use_ffmpeg:
      self.P(f"[{stream.name}] Opening capture with CAP_FFMPEG...")
      cap = cv2.VideoCapture(stream.url, cv2.CAP_FFMPEG)
    else:
      cap = cv2.VideoCapture(stream.url)

    cap_prop_buff = self._get_stream_cfg_value(stream, "CAP_PROP_BUFFERSIZE", self.cfg_cap_prop_buffersize)
    if cap_prop_buff and cap_prop_buff > 0:
      cap.set(cv2.CAP_PROP_BUFFERSIZE, cap_prop_buff)
    return cap

  def _get_stream_cfg_value(self, stream: _StreamContext, key: str, default=None):
    if key in stream.config and stream.config[key] is not None:
      return stream.config[key]
    attr_name = f"cfg_{key.lower()}"
    if hasattr(self, attr_name):
      return getattr(self, attr_name)
    return self.config.get(key, default)

  def _get_streams_snapshot(self) -> Iterable[_StreamContext]:
    with self._streams_lock:
      return list(self._streams.values())

  def _run_data_aquisition_step(self):
    inputs = []
    for stream in self._get_streams_snapshot():
      if not stream.active or not stream.has_connection:
        continue

      with stream.lock:
        capture = stream.capture
      if capture is None or not capture.isOpened():
        self.P(f"[{stream.name}] Capture closed mid-loop. Scheduling reconnect.", color="r")
        self._mark_stream_disconnected(stream, force_reconnect=True)
        continue

      has_frame, frame = self._capture_read(stream, capture)
      if not has_frame or frame is None:
        self.P(f"[{stream.name}] Failed to read frame. Scheduling reconnect.", color="r")
        self._mark_stream_disconnected(stream, force_reconnect=True)
        continue

      frame = self._maybe_resize_crop_stream(stream, frame)
      frame = np.ascontiguousarray(frame[:, :, ::-1])

      self._verify_configured_size(stream, frame)

      stream.crt_frame += 1
      stream.metadata.frame_current = stream.crt_frame

      inputs.append(self._new_input(img=frame, metadata=self._build_frame_metadata(stream)))

    if inputs:
      self._add_inputs(inputs)
      self.has_connection = True
    elif not any(ctx.has_connection for ctx in self._streams.values()):
      self.has_connection = False
    return

  def _verify_configured_size(self, stream: _StreamContext, frame):
    configured_h = self._get_stream_cfg_value(stream, "CONFIGURED_H", self.cfg_configured_h)
    configured_w = self._get_stream_cfg_value(stream, "CONFIGURED_W", self.cfg_configured_w)

    mismatch_h = configured_h > 0 and configured_h != frame.shape[0]
    mismatch_w = configured_w > 0 and configured_w != frame.shape[1]
    if (mismatch_h or mismatch_w) and not stream.configured_size_error:
      stream.configured_size_error = True
      msg = (
        f"ERROR: Stream '{stream.name}' configured at {configured_h}x{configured_w} "
        f"decoded at {frame.shape[0]}x{frame.shape[1]}"
      )
      self._create_notification(
        notif=ct.STATUS_TYPE.STATUS_ABNORMAL_FUNCTIONING,
        stream_name=stream.name,
        msg=msg,
      )
    return

  def _build_frame_metadata(self, stream: _StreamContext):
    data = self.deepcopy(stream.metadata.__dict__)
    data.update(
      source_name=stream.name,
      source_index=stream.index,
      source_url=stream.url,
      connected=stream.has_connection,
    )
    return data

  def _maybe_resize_crop_stream(self, stream: _StreamContext, img):
    if stream.resize_hw is not None:
      target_h, target_w = stream.resize_hw
      self.start_timer(self._stream_timer_id(stream, "cv2_resize"))
      img = cv2.resize(img, dsize=(target_w, target_h))
      self.end_timer(self._stream_timer_id(stream, "cv2_resize"))
    elif stream.frame_crop is not None:
      self.start_timer(self._stream_timer_id(stream, "cv2_crop"))
      top, left, bottom, right = stream.frame_crop
      img = img[top:bottom, left:right]
      self.end_timer(self._stream_timer_id(stream, "cv2_crop"))
    return img

  def _configure_stream_resize_crop(self, stream: _StreamContext, height: int, width: int):
    stream.resize_hw = None
    stream.frame_crop = None

    metadata_cfg = {}
    base_cfg = self.cfg_stream_config_metadata or {}
    if isinstance(base_cfg, dict):
      metadata_cfg.update(base_cfg)
    stream_cfg = stream.config.get("STREAM_CONFIG_METADATA", {}) or {}
    if isinstance(stream_cfg, dict):
      metadata_cfg.update(stream_cfg)

    resize_h = metadata_cfg.get(ct.RESIZE_H)
    resize_w = metadata_cfg.get(ct.RESIZE_W)
    if resize_h is not None and resize_h > 0:
      ratio = height / width if width else 1
      resize_w = resize_w if resize_w is not None else int(resize_h / ratio)
      stream.resize_hw = (int(resize_h), int(resize_w))
      self.P(
        f"[{stream.name}] Live resize enabled from {(height, width)} to {stream.resize_hw}",
        color="y",
      )
      stream.metadata.resize_hw = stream.resize_hw

    frame_crop = metadata_cfg.get(ct.FRAME_CROP, metadata_cfg.get(ct.CAPTURE_CROP))
    if frame_crop is not None and stream.resize_hw is None:
      if (
        isinstance(frame_crop, list)
        and len(frame_crop) == 4
        and frame_crop[0] <= frame_crop[2]
        and frame_crop[1] <= frame_crop[3]
      ):
        stream.frame_crop = frame_crop
        cropped_h = frame_crop[2] - frame_crop[0]
        cropped_w = frame_crop[3] - frame_crop[1]
        stream.metadata.frame_h = cropped_h
        stream.metadata.frame_w = cropped_w
        self.P(
          f"[{stream.name}] Enabling cropping from {(height, width)} to {(cropped_h, cropped_w)} "
          f"(TLBR:{frame_crop})",
          color="y",
        )
      else:
        self.P(
          f"[{stream.name}] Invalid FRAME_CROP {frame_crop}. Using original frame size.",
          color="r",
        )
    return

  def _capture_read(self, stream: _StreamContext, capture):
    has_frame, frame = False, None
    timer_suffix = self._stream_timer_id(stream, "cv2")

    if ((self.cap_resolution >= self._get_stream_fps_max_thr(stream) and self.cfg_amd_target_dps == 0)
        or self.is_intel()):
      self.start_timer(f"{timer_suffix}_read")
      self.start_timer(f"{timer_suffix}_read_dps{self.cap_resolution}")
      has_frame, frame = capture.read()
      self.end_timer(f"{timer_suffix}_read_dps{self.cap_resolution}")
      self.end_timer(f"{timer_suffix}_read")
    else:
      nr_grabs = self._get_nr_grabs(stream)
      self.start_timer(f"{timer_suffix}_grab_retrv_dps{self.cap_resolution}")
      self.start_timer(f"{timer_suffix}_grab_x{nr_grabs}")
      for _ in range(nr_grabs):
        self.start_timer(f"{timer_suffix}_grab")
        try:
          _ = capture.grab()
        except Exception:
          self.end_timer(f"{timer_suffix}_grab")
          break
        self.end_timer(f"{timer_suffix}_grab")
      self.end_timer(f"{timer_suffix}_grab_x{nr_grabs}")
      self.start_timer(f"{timer_suffix}_retrieve")
      has_frame, frame = capture.retrieve()
      self.end_timer(f"{timer_suffix}_retrieve")
      self.end_timer(f"{timer_suffix}_grab_retrv_dps{self.cap_resolution}")

    if not self.is_intel() and self.cfg_amd_target_dps == 0:
      self._recalc_cap_resolution(stream)
    return has_frame, frame

  def _stream_timer_id(self, stream: _StreamContext, base: str) -> str:
    suffix = self.sanitize_name(stream.name)
    return f"{base}__{suffix}"

  def _get_stream_fps_max_thr(self, stream: _StreamContext):
    fps = stream.metadata.__dict__.get("fps")
    if fps is None:
      return float("inf")
    return fps * 0.6

  def _get_nr_grabs(self, stream: _StreamContext):
    if self.cfg_amd_target_dps > 0:
      n_grabs = int(np.ceil(self.cfg_cap_resolution / self.cfg_amd_target_dps))
    else:
      fps = stream.metadata.__dict__.get("fps", 1)
      n_grabs = int(np.ceil(fps / self.cap_resolution)) + 1
    return n_grabs

  def _recalc_cap_resolution(self, stream: _StreamContext):
    fps = stream.metadata.__dict__.get("fps")
    if fps is None:
      return
    WARM_UP_SEC = 60
    MAX_TIME = 2
    if self.cap_resolution >= fps * 0.6:
      nr_streams = self.get_nr_parallel_captures() + len(self._streams)
      timer_id = f"{self._stream_timer_id(stream, 'cv2')}_read"
      read_stats = self.get_timer(timer_id)
      if read_stats is None:
        return
      read_time = read_stats.get("MEAN", 0)
      read_count = read_stats.get("COUNT", 0)
      if read_count > (self.cap_resolution * WARM_UP_SEC):
        total_read_time = read_time * self.cap_resolution * nr_streams
        if total_read_time >= MAX_TIME:
          configured_cap_res = self.get_cap_or_forced_resolution()
          self._heuristic_cap_resolution = round(MAX_TIME / (read_time * nr_streams), 1)
          self.P(
            "Total cv2r time @{} load: {:.4f}s({:.4f}s/strm/itr) exceding {}s. "
            "Forcing cap {} @ {:.1f} dps, nr_grabs={} (stream fps: {} cap dps:{}/{})".format(
              nr_streams,
              total_read_time,
              read_time,
              MAX_TIME,
              self.cfg_name,
              self._heuristic_cap_resolution,
              self._get_nr_grabs(stream),
              stream.metadata.__dict__.get("fps"),
              self.cap_resolution,
              configured_cap_res,
            ),
            color="m",
          )
    return

  def _release_stream_capture(self, stream: _StreamContext):
    with stream.lock:
      if stream.capture is not None:
        try:
          stream.capture.release()
        except Exception:
          pass
        stream.capture = None
      stream.has_connection = False
    return

  def _mark_stream_disconnected(self, stream: _StreamContext, force_reconnect: bool):
    self._release_stream_capture(stream)
    if not self.is_reconnectable:
      stream.active = False
    if force_reconnect and self.is_reconnectable:
      self._launch_reconnect(stream)
    self._update_global_connection_flags()
    return

  def _update_global_metadata(self):
    streams_snapshot = {
      name: self.deepcopy(ctx.metadata.__dict__)
      for name, ctx in self._streams.items()
    }
    active_names = [name for name, ctx in self._streams.items() if ctx.has_connection]
    self._metadata.update(streams=streams_snapshot, connected_streams=active_names)
    self._stream_metadata.streams = streams_snapshot
    self._stream_metadata.connected_streams = active_names
    return

  def _update_global_connection_flags(self):
    total_issues = sum(ctx.nr_connection_issues for ctx in self._streams.values())
    self.nr_connection_issues = total_issues
    self.has_connection = any(ctx.has_connection for ctx in self._streams.values())
    return



import sys
import types
import unittest

from naeural_core.serving.base import basic_th
from naeural_core.serving.mixins_base import trt_mixin as trt_mixin_module


class _FallbackHarness(basic_th.UnifiedFirstStage):
  def __init__(self):
    self._messages = []
    self._environment_variables = {}
    self._upstream_config = {}
    self.cfg_allow_backend_fallback = True
    self.cfg_strict_backend = False
    self.server_name = "TEST_BACKEND"
    self._platform_backend_priority = []
    return

  def P(self, msg, color=None, **kwargs):  # pylint: disable=unused-argument
    self._messages.append(msg)
    return


class _TrtSkipHarness(trt_mixin_module.TensortRTMixin):
  def __init__(self):
    self._messages = []
    self.server_name = "TEST_TRT_SKIP"
    self.version = "0.0.0"
    self.dev = 0
    self.cfg_fp16 = False
    self.cfg_model_trt_filename = "model.onnx"
    self.cfg_trt_url = "minio:dummy/model.onnx"
    return

  def P(self, msg, color=None, **kwargs):  # pylint: disable=unused-argument
    self._messages.append(msg)
    return

  def download_model_for_backend(self, *args, **kwargs):  # pylint: disable=unused-argument
    raise AssertionError("TRT compatibility guard should skip before downloads")


class _FakeDeviceProps:
  major = 12
  minor = 0


class _FakeCuda:
  @staticmethod
  def is_available():
    return True

  @staticmethod
  def device_count():
    return 1

  @staticmethod
  def current_device():
    return 0

  @staticmethod
  def get_device_properties(_idx):
    return _FakeDeviceProps()


class _FakeTorch:
  cuda = _FakeCuda

  class device:
    def __init__(self, index=None):
      self.index = index


class TestBackendFallback(unittest.TestCase):
  def test_prepare_model_falls_back_to_ths(self):
    harness = _FallbackHarness()
    harness._platform_backend_priority = ["trt", "ths"]

    def _prepare_trt_model(*args, **kwargs):  # pylint: disable=unused-argument
      raise RuntimeError("TRT build failed")

    def _prepare_ts_model(*args, **kwargs):  # pylint: disable=unused-argument
      return "ths-model", {"ok": True}

    harness._prepare_trt_model = _prepare_trt_model
    harness._prepare_ts_model = _prepare_ts_model

    backend_map = {
      "trt": ("model_trt.onnx", "minio:dummy/model_trt.onnx"),
      "ths": ("model.ths", "minio:dummy/model.ths"),
    }

    model, _config, _fn = harness.prepare_model(
      backend_model_map=backend_map,
      return_config=True
    )
    self.assertEqual(model, "ths-model")
    self.assertIn("falling back to ths", " ".join(harness._messages).lower())

  def test_prepare_model_strict_backend_raises(self):
    harness = _FallbackHarness()
    harness._platform_backend_priority = ["trt", "ths"]

    def _prepare_trt_model(*args, **kwargs):  # pylint: disable=unused-argument
      raise RuntimeError("TRT build failed")

    harness._prepare_trt_model = _prepare_trt_model

    backend_map = {
      "trt": ("model_trt.onnx", "minio:dummy/model_trt.onnx"),
    }

    with self.assertRaises(RuntimeError):
      harness.prepare_model(
        backend_model_map=backend_map,
        forced_backend="trt",
        return_config=True
      )

  def test_trt_skips_unsupported_sm(self):
    fake_trt = types.ModuleType("tensorrt")
    fake_trt.__version__ = "8.6.1"

    original_trt = sys.modules.get("tensorrt")
    original_th = trt_mixin_module.th
    sys.modules["tensorrt"] = fake_trt
    trt_mixin_module.th = _FakeTorch()
    try:
      harness = _TrtSkipHarness()
      result = harness._prepare_trt_model(
        url="minio:dummy/model_trt.onnx",
        fn_model="model_trt.onnx",
        return_config=True,
        batch_size=1,
        allow_backend_fallback=True,
        strict_backend=False
      )
      self.assertEqual(result, (None, None))
      self.assertIn("does not support sm", harness._last_backend_error.lower())
    finally:
      if original_trt is None:
        sys.modules.pop("tensorrt", None)
      else:
        sys.modules["tensorrt"] = original_trt
      trt_mixin_module.th = original_th





class _BasePluginAPIMixin:
  def __init__(self) -> None:
    super(_BasePluginAPIMixin, self).__init__()
    
    self.__chain_state_initialized = False
    return
  
  # Obsolete
  def _pre_process(self):
    """
    Called before process. Currently (partially) obsolete

    Returns
    -------
    TBD.

    """
    return
  
  def _post_process(self):
    """
    Called after process. Currently (partially) obsolete

    Returns
    -------
    TBD.

    """
    return
  
  
  def step(self):
    """
    The main code of the plugin (loop iteration code). Called at each iteration of the plugin loop.

    Returns
    -------
    None.

    """
    return
  
  
  def process(self):
    """
    The main code of the plugin (loop iteration code). Called at each iteration of the plugin loop.

    Returns
    -------
    Payload.

    """
    return self.step()
  
  def _process(self):
    """
    The main code of the plugin (loop iteration code.

    Returns
    -------
    Payload.

    """
    return self.process()

  
  def on_init(self):
    """
    Called at init time in the plugin thread.

    Returns
    -------
    None.

    """      
    return
  
  def _on_init(self):
    """
    Called at init time in the plugin thread.

    Returns
    -------
    None.

    """
    self.P("Default plugin `_on_init` called for plugin initialization...")
    self.on_init()
    return


  def on_close(self):
    """
    Called at shutdown time in the plugin thread.

    Returns
    -------
    None.

    """      
    return


  def _on_close(self):
    """
    Called at shutdown time in the plugin thread.

    Returns
    -------
    None.

    """
    self.P("Default plugin `_on_close` called for plugin cleanup at shutdown...")
    self.maybe_archive_upload_last_files()
    # Auto-cleanup semaphore if configured
    self._semaphore_auto_cleanup()
    self.on_close()
    return

  def on_command(self, data, **kwargs):
    """
    Called when a INSTANCE_COMMAND is received by the plugin instance.
    
    The command is sent via `cmdapi_send_instance_command` as in below simplified example:
    
    ```python
      pipeline = "some_app_pipeline"
      signature = "CONTAINER_APP_RUNNER"
      instance_id = "CONTAINER_APP_1e8dac"
      node_address = "0xai_1asdfG11sammamssdjjaggxffaffaheASSsa"
      
      instance_command = "RESTART"
      
      plugin.cmdapi_send_instance_command(
        pipeline=pipeline,
        signature=signature,
        instance_id=instance_id,
        instance_command=instance_command,
        node_address=node_address,
      )
    ```
    
    while the `on_command` method should look like this:
    
    ```python
      def on_command(self, data, **kwargs):
        if data == "RESTART":
          self.P("Restarting ...")
        elif data == "STOP":
          self.P("Stopping ...")
        else:
          self.P(f"Unknown command: {data}")
        return
    ```
    
    The `instance_command` is passed to this method as `data` and in fact can be a dict with extra data. If
    `instance_command` contains `COMMAND_PARAMS` dict then all the key-value pairs in the `COMMAND_PARAM` dict 
    will be passed as kwargs to this method - see below example:
    
    ```python
      instance_command = {
        "COMMAND_PARAMS": {
          "some_param1": "value1",
          "some_param2": "value2",
        }
      }
      
      plugin.cmdapi_send_instance_command(
        pipeline=pipeline,
        signature=signature,
        instance_id=instance_id,
        instance_command=instance_command,
        node_address=node_address,
      )
    ```
    Then this `on_command` method should look like this:
    
    ```python
    
      def on_command(self, data, some_param1=None, some_param2=None, **kwargs):
        if some_param1:
          self.P(f"Received some_param1: {some_param1}")
        if some_param2:
          self.P(f"Received some_param2: {some_param2}")
        # Process the command here
        return
    ```

    Parameters
    ----------
    data : any
      object, string, etc.

    Returns
    -------
    None.

    """
    return

  def _on_command(self, data, default_configuration=None, current_configuration=None, **kwargs):
    """
    Called when a INSTANCE_COMMAND is received by the plugin instance.
    
    The command is sent via `cmdapi_send_instance_command` as in below simplified example:
    
    ```python
      pipeline = "some_app_pipeline"
      signature = "CONTAINER_APP_RUNNER"
      instance_id = "CONTAINER_APP_1e8dac"
      node_address = "0xai_1asdfG11sammamssdjjaggxffaffaheASSsa"
      
      instance_command = "RESTART"
      
      plugin.cmdapi_send_instance_command(
        pipeline=pipeline,
        signature=signature,
        instance_id=instance_id,
        instance_command=instance_command,
        node_address=node_address,
      )
    ```
    
    while the `on_command` method should look like this:
    
    ```python
      def on_command(self, data, **kwargs):
        if data == "RESTART":
          self.P("Restarting ...")
        elif data == "STOP":
          self.P("Stopping ...")
        else:
          self.P(f"Unknown command: {data}")
        return
    ```
    
    The `instance_command` is passed to this method as `data` and in fact can be a dict with extra data. If
    `instance_command` contains `COMMAND_PARAMS` dict then all the key-value pairs in the `COMMAND_PARAM` dict 
    will be passed as kwargs to this method - see below example:
    
    ```python
      instance_command = {
        "COMMAND_PARAMS": {
          "some_param1": "value1",
          "some_param2": "value2",
        }
      }
      
      plugin.cmdapi_send_instance_command(
        pipeline=pipeline,
        signature=signature,
        instance_id=instance_id,
        instance_command=instance_command,
        node_address=node_address,
      )
    ```
    Then this `on_command` method should look like this:
    
    ```python
    
      def on_command(self, data, some_param1=None, some_param2=None, **kwargs):
        if some_param1:
          self.P(f"Received some_param1: {some_param1}")
        if some_param2:
          self.P(f"Received some_param2: {some_param2}")
        # Process the command here
        return
    ```
    
    Parameters
    ----------
    data : any
      object, string, etc.

    Returns
    -------
    None.

    """
    self.P("Default plugin `_on_command`...")

    if (isinstance(data, str) and data.upper() == 'DEFAULT_CONFIGURATION') or default_configuration:
      self.P("Received \"DEFAULT_CONFIGURATION\" command...")
      self.add_payload_by_fields(
        default_configuration=self._default_config,
        command_params=data,
      )
      return
    if (isinstance(data, str) and data.upper() == 'CURRENT_CONFIGURATION') or current_configuration:
      self.P("Received \"CURRENT_CONFIGURATION\" command...")
      self.add_payload_by_fields(
        current_configuration=self._upstream_config,
        command_params=data,
      )
      return

    self.on_command(data, **kwargs)
    return


  def _on_config(self):
    """
    Called when the instance has just been reconfigured

    Parameters
    ----------
    None

    Returns
    -------
    None.

    """
    self.P("Default plugin {} `_on_config` called...".format(self.__class__.__name__))
    if hasattr(self, 'on_config'):
      self.on_config()
    return


  ###
  ### Chain State
  ### 
  
  def __chainstorage_memory(self):
    return self.plugins_shmem
  
  def __maybe_wait_for_chain_state_init(self):
    # TODO: raise exception if not found after a while

    while not self.plugins_shmem.get('__chain_storage_set'):
      self.sleep(0.1)
    
    if not self.__chain_state_initialized:
      self.P(" ==== Chain state initialized.")
    self.__chain_state_initialized = True
    return
  
  def chainstore_set(self, key, value, readonly=False, token=None, debug=False, extra_peers=[]):
    """
    Set data in the R1 Chain Storage.

    This method stores a key-value pair in the distributed chain storage, broadcasting
    the data to peer nodes for replication and waiting for confirmations.

    IMPORTANT - JSON Normalization:
    -------------------------------
    Values are automatically normalized through JSON serialization before storage.
    This ensures deterministic behavior across the distributed network, as all nodes
    will store and compare identical data structures regardless of the original
    Python types used.

    Key implications of JSON normalization:
      - Dictionary keys are converted to strings (e.g., {8080: "http"} becomes {"8080": "http"})
      - Values must be JSON-serializable (no datetime, bytes, custom objects without serialization)
      - Numeric precision follows JSON spec (integers preserved, floats may have precision limits)
      - Dictionary ordering is preserved (Python 3.7+)

    Example:
    --------
    ```python
    # Integer keys are converted to string keys
    data = {'ports': {8080: 'http', 443: 'https'}}
    self.chainstore_set('my_key', data)

    # When retrieved, keys will be strings:
    result = self.chainstore_get('my_key')
    # result = {'ports': {'8080': 'http', '443': 'https'}}
    ```

    Parameters
    ----------
    key : str
      The key under which to store the value. Should be a unique identifier
      within the chain storage namespace.

    value : any (JSON-serializable)
      The value to store. Must be JSON-serializable. Complex types like
      datetime objects should be converted to strings before storing.

    readonly : bool, optional
      If True, the key-value pair becomes read-only and cannot be overwritten
      by other owners. Default is False.

    token : any, optional
      A token for access control. If provided, subsequent read/write operations
      on this key must provide the same token. Default is None.

    debug : bool, optional
      If True, enables verbose logging of the operation including peer counts,
      confirmation status, and timing information. Default is False.

    extra_peers : list, optional
      Additional peer addresses to broadcast the data to, beyond the default
      chain peers. Default is an empty list.

    Returns
    -------
    bool
      True if the value was successfully stored and confirmed by peers,
      False if the operation failed (timeout, no confirmations, etc.)

    See Also
    --------
    chainstore_get : Retrieve a value from chain storage
    chainstore_hset : Store a value in a hash set within chain storage
    """
    self.start_timer("chainstore_set")
    memory = self.__chainstorage_memory()
    result = False
    try:
      self.__maybe_wait_for_chain_state_init()
      func = memory.get('__chain_storage_set')

      # JSON normalization: ensure deterministic storage by normalizing through JSON.
      # This converts integer dict keys to strings and ensures consistent data structures
      # across all nodes in the distributed network. Without this normalization, the
      # local storage could have {8080: "http"} while the transmitted/received data
      # would have {"8080": "http"}, causing confirmation comparisons to fail.
      value = self.json_loads(self.json_dumps(value))

      specific_peers = self.cfg_chainstore_peers or []
      if isinstance(specific_peers, str):
        specific_peers = [specific_peers]
      elif not isinstance(specific_peers, list):
        specific_peers = []
      if isinstance(extra_peers, list):
        specific_peers += extra_peers
      # filter self address from specific_peers  
      specific_peers = [x for x in specific_peers if x != self.ee_addr]
      
      if func is not None:
        if debug:
          self.P("Setting data: {} -> {}".format(key, value), color="green")
        result = func(
          key, value, 
          readonly=readonly, token=token, peers=specific_peers, debug=debug
        )
        elapsed = self.end_timer("chainstore_set")        
        if debug:
          self.P(" ====> `chainstore_set`in {:.4f}s".format(elapsed), color="green")
      else:
        elapsed = self.end_timer("chainstore_set")
        if debug:
          self.P("No chain storage set function found in {:.4f}s".format(elapsed), color="red")
    except Exception as ex:
      elapsed = self.end_timer("chainstore_set")
      msg = "Error in chainstore_set: {} after {:.4f}s".format(ex, elapsed)
      self.P(msg, color="red")      
    return result
  
  
  def chainstore_get(self, key, token=None, debug=False):
    """
    Get data from the R1 Chain Storage
    
    Parameters
    ----------
    key : str
      Key
      
    token : any, optional
      Token, by default None
    """
    self.start_timer("chainstore_get")
    memory = self.__chainstorage_memory()
    self.__maybe_wait_for_chain_state_init()
    value = None
    msg = ""
    try:
      start_search = self.time()
      found = True
      while memory.get('__chain_storage_get') is None:
        self.sleep(0.1)
        if self.time() - start_search > 10:
          msg = "Error: chain storage get function not found after 10 seconds"
          self.P(msg, color="red")
          found = False
          break
      func = memory.get('__chain_storage_get')
      if func is not None:
        value = func(key, token=token, debug=debug)
        elapsed = self.end_timer("chainstore_get")
        if debug:
          self.P("====> `chainstore_get`: {} -> {} in {:.4f}s".format(key, value, elapsed))
      else:
        elapsed = self.end_timer("chainstore_get")
        if debug:
          self.P("No chain storage get function found in {:.4f}s".format(elapsed), color="red")
    except Exception as ex:
      elapsed = self.end_timer("chainstore_get")  
      msg = "Error in chainstore_get: {} after {:.4f}s".format(ex, elapsed)
      self.P(msg, color="red")
    return value

  
  def __hset_index(self, hkey):
    hkey_hash = self.get_hash(hkey, algorithm='sha256', length=10)
    return f"hs:{hkey_hash}:"

  
  def __hset_key(self, hkey, key):
    b64key = self.str_to_base64(key, url_safe=True)
    return self.__hset_index(hkey) + b64key

  
  def chainstore_hget(self, hkey, key, token=None, debug=False):
    """    
    This is a basic implementation of a hash get operation in the chain storage.
    It uses a hash-based string composition to create a composed key.
    """
    start_1 = self.time()
    composed_key = self.__hset_key(hkey, key)
    elapsed_1 = self.time() - start_1
    start_2 = self.time()
    result = self.chainstore_get(composed_key, token=token, debug=debug)
    elapsed_2 = self.time() - start_2
    if debug:
      self.P(f"HGET: '{composed_key}' (index_time={elapsed_1:.4f}s, get_time={elapsed_2:.4f}s)")
    return result  

  def chainstore_hset(self, hkey, key, value, readonly=False, token=None, debug=False, extra_peers=[]):
    """
    This is a basic implementation of a hash set operation in the chain storage.
    It uses a hash-based string composition to create a composed key.
    """
    start_1 = self.time()
    composed_key = self.__hset_key(hkey, key)
    elapsed_1 = self.time() - start_1
    start_2 = self.time()
    result = self.chainstore_set(composed_key, value, readonly=readonly, token=token, debug=debug, extra_peers=extra_peers)
    elapsed_2 = self.time() - start_2
    if debug:
      self.P(f"HSET: '{composed_key}' (index_time={elapsed_1:.4f}s, set_time={elapsed_2:.4f}s)")
    return result

  
  def chainstore_hlist(self, hkey : str, token=None, debug=False):
    index = self.__hset_index(hkey)
    memory = self.__chainstorage_memory()
    self.__maybe_wait_for_chain_state_init()
    chain_storage = memory.get('__chain_storage')
    result = []
    for key in chain_storage:
      if not isinstance(key, str) or key == "":
        self.P("Invalid key type in chain storage: {}".format(type(key)), color="red")
        continue
      if key.startswith(index):
        b64field = key[len(index):]
        field = self.base64_to_str(b64field, url_safe=True)
        result.append(field)
      #end if 
    #end for
    return result

  
  def chainstore_hkeys(self, hkey : str, token=None, debug=False):
    return self.chainstore_hlist(hkey, token=token, debug=debug)


  def chainstore_hgetall(self, hkey : str, token=None, debug=False):
    keys = self.chainstore_hlist(hkey, token=token, debug=debug)
    result = {}
    for key in keys:
      value = self.chainstore_hget(hkey, key, token=token, debug=debug)
      result[key] = value
    return result
    
  
  # # @property
  # # This CANNOT be a property, as it can be a blocking operation.
  # def _chainstorage(self): # TODO: hide/move/protect this
  #   self.__maybe_wait_for_chain_state_init()
  #   return self.plugins_shmem.get('__chain_storage')

  
  def get_instance_path(self):
    return [self.ee_id, self._stream_id, self._signature, self.cfg_instance_id]  
  
  ###
  ### END Chain State
  ###
  
    


"""


"""

from naeural_core.business.base import BasePluginExecutor as BaseClass
from naeural_core.business.mixins_libs.network_processor_mixin import _NetworkProcessorMixin, __VER__

_CONFIG = {
  **BaseClass.CONFIG,
  
  'ALLOW_EMPTY_INPUTS' : False,

  'MAX_INPUTS_QUEUE_SIZE' : 1024,

  "PROCESS_DELAY" : 0,  # seconds, how often to process the network
  "LOG_STATS_PERIOD": 5 * 60,  # seconds, how often to log how many messages were processed from each signature

  "NETWORK_ROUTE_BY_HANDLER": True,
  
  'VALIDATION_RULES' : {
    **BaseClass.CONFIG['VALIDATION_RULES'],
    'MAX_INPUTS_QUEUE_SIZE': {
      'DESCRIPTION': "Advanced setting: Controls the queue size of the plugin instance. In the case of NetworkProcessor plugins this queue should be allowed to be much larger due to the reduced size of the messages and the fast-paced nature of them.",
      'TYPE': 'int',
      'MIN_VAL': 1,
      'MAX_VAL': 4096,
    },
    'NETWORK_ROUTE_BY_HANDLER': {
      'DESCRIPTION': "When true, drop payloads that do not have a registered handler before verification/dispatch.",
      'TYPE': 'bool',
    },
  },  
}

class NetworkProcessorPlugin(
  BaseClass,
  _NetworkProcessorMixin
):
  CONFIG = _CONFIG

  def _on_init(self):
    self.network_processor_init()
    self.P(
      "NetworkProcessorPlugin v{} base initialization completed. Proceeding to custom init...".format(__VER__),
      color="green"
    )
    self.on_init()
    return


  def _process(self):
    """
    This method must be protected while the child plugins should have normal `process`
    """
    self.network_processor_loop()
    return self.process()



"""
Semaphored Paired Plugin Mixin

This mixin provides semaphore-based synchronization between paired plugins,
enabling coordination between native plugins (providers) and Container App
Runners (consumers).

Use Cases:
  - Keysoft: Jeeves FastAPI (native) + CAR container
  - RedMesh: Pentester API (native) + CAR UI container
  - CerviGuard: Local Serving API (native) + WAR container

Provider plugins (native) use:
  - semaphore_set_env(): Set environment variables for paired plugins
  - semaphore_set_ready(): Signal initialization complete
  - semaphore_clear(): Clean up on shutdown

Consumer plugins (CAR/WAR) use:
  - semaphore_is_ready(): Check if all dependencies are ready
  - semaphore_get_env(): Collect environment variables from dependencies
  - semaphore_get_missing(): Get list of missing semaphores

Configuration:
  Provider: SEMAPHORE = "UNIQUE_KEY"
  Consumer: SEMAPHORED_KEYS = ["KEY1", "KEY2"]
"""

from time import time as tm


class _SemaphoredPairedPluginMixin(object):
  """
  Mixin for coordinating startup and environment exchange between paired plugins
  (e.g., a native plugin and a Container App Runner) using shared memory semaphores.
  """

  def __init__(self):
    """
    Initialize all semaphore-related state variables.

    Consumer state (for plugins waiting on dependencies):
      - __semaphore_wait_start: Timestamp when waiting began
      - __semaphore_ready_logged: Set of semaphores already logged as ready

    Provider state (for plugins signaling readiness):
      - _semaphore_signaled: Flag to prevent duplicate signaling
    """
    # Consumer state (existing - for plugins that wait for others)
    self.__semaphore_wait_start = None
    self.__semaphore_ready_logged = set()

    # Provider state (new - for plugins that signal they're ready)
    self._semaphore_signaled = False

    super(_SemaphoredPairedPluginMixin, self).__init__()
    return

  def _semaphore_Pd(self, msg):
    """Safe debug logging - only logs if Pd method is available."""
    if hasattr(self, 'Pd') and callable(self.Pd):
      self.Pd(msg)
    return

  @staticmethod
  def _semaphore_prefixed_key(semaphore_key, env_key):
    """Build a prefixed env key from semaphore key and env key."""
    return "{}_{}".format(semaphore_key, env_key)

  # ============================================================================
  # Provider Methods (for native plugins that signal readiness)
  # ============================================================================

  def _semaphore_ensure_structure(self):
    """Ensure the semaphore data structure exists in shared memory."""
    semaphore_key = getattr(self, 'cfg_semaphore', None)
    if not semaphore_key:
      return None

    if semaphore_key not in self.plugins_shmem:
      self.plugins_shmem[semaphore_key] = {
        'is_ready': False,
        'env': {},
        'metadata': {
          'instance_id': self.cfg_instance_id,
          'plugin_signature': self.__class__.__name__,
          'ready_timestamp': None,
        }
      }
    return self.plugins_shmem[semaphore_key]

  def semaphore_set_ready(self):
    """
    Signal that this plugin is ready.
    Sets 'is_ready' = True in the shared memory segment identified by cfg_semaphore.

    Returns
    -------
    bool
      True if semaphore was set, False if SEMAPHORE not configured
    """
    semaphore_key = getattr(self, 'cfg_semaphore', None)
    if not semaphore_key:
      return False

    semaphore_data = self._semaphore_ensure_structure()
    semaphore_data['is_ready'] = True
    semaphore_data['metadata']['ready_timestamp'] = tm()
    self._semaphore_Pd("Semaphore '{}' set to READY".format(semaphore_key))
    return True

  def semaphore_set_env(self, key, value):
    """
    Set an environment variable to be shared with the paired plugin.

    The key is stored twice: once prefixed with the semaphore name for namespacing,
    and once without prefix for consumers that expect the raw key.

    Parameters
    ----------
    key : str
      The environment variable name (stored as both prefixed and raw)
    value : any
      The environment variable value (will be converted to string)

    Returns
    -------
    bool
      True if env var was set, False if SEMAPHORE not configured
    """
    semaphore_key = getattr(self, 'cfg_semaphore', None)
    if not semaphore_key:
      return False

    semaphore_data = self._semaphore_ensure_structure()

    # Store both prefixed and raw variants for flexibility
    value_str = str(value)
    prefixed_key = self._semaphore_prefixed_key(semaphore_key, key)
    semaphore_data['env'][prefixed_key] = value_str
    semaphore_data['env'][key] = value_str
    self._semaphore_Pd(
      "Semaphore '{}' env vars set: {} / {} = {}".format(
        semaphore_key, prefixed_key, key, value))
    return True

  def semaphore_set_env_dict(self, env_dict):
    """
    Set multiple environment variables at once.

    Each key is stored with both prefixed and raw variants, matching
    semaphore_set_env() behavior.

    Parameters
    ----------
    env_dict : dict
      Dictionary of {key: value} pairs.

    Returns
    -------
    bool
      True if all env vars were set, False if SEMAPHORE not configured
    """
    semaphore_key = getattr(self, 'cfg_semaphore', None)
    if not semaphore_key:
      return False

    for key, value in env_dict.items():
      self.semaphore_set_env(key, value)
    return True

  def semaphore_clear(self):
    """
    Clear the semaphore (e.g., on plugin shutdown).

    This signals to waiting plugins that this dependency is no longer available.
    Should be called in on_close() of provider plugins.
    """
    semaphore_key = getattr(self, 'cfg_semaphore', None)
    if not semaphore_key:
      return

    if semaphore_key in self.plugins_shmem:
      self.plugins_shmem[semaphore_key]['is_ready'] = False
      self.plugins_shmem[semaphore_key]['metadata']['ready_timestamp'] = None
      self._semaphore_Pd("Semaphore '{}' cleared".format(semaphore_key))
    return


  def _semaphore_maybe_auto_signal(self):
    """
    Automatically signal semaphore readiness when conditions are met.

    This method is called automatically from the plugin's process loop.
    It checks if the plugin is ready and signals the semaphore accordingly.

    Readiness is determined by is_plugin_ready() from _PluginReadinessMixin:
      - None: use default (uvicorn_server_started or _init_process_finalized)
      - False: explicitly deferred, wait for set_plugin_ready(True)
      - True: explicitly ready

    Once ready, this method:
      1. Ensures semaphore structure exists
      2. Calls _setup_semaphore_env() hook (if implemented)
      3. Signals readiness via semaphore_set_ready()
      4. Sets _semaphore_signaled flag to prevent duplicate signaling

    Returns
    -------
    None
    """
    # Early exit if semaphore not configured
    semaphore_key = getattr(self, 'cfg_semaphore', None)
    if not semaphore_key:
      return

    # Early exit if already signaled
    if self._semaphore_signaled:
      return

    # Check unified readiness (from _PluginReadinessMixin)
    is_ready = self.is_plugin_ready()

    # Early exit if not ready yet
    if not is_ready:
      return

    # Plugin is ready - signal semaphore
    try:
      # Ensure semaphore structure exists
      self._semaphore_ensure_structure()

      # Call hook method if implemented (for plugin-specific env vars)
      if hasattr(self, '_setup_semaphore_env') and callable(self._setup_semaphore_env):
        self._setup_semaphore_env()

      # Signal readiness
      self.semaphore_set_ready()

      # Mark as signaled to prevent duplicate signaling
      self._semaphore_signaled = True

      self._semaphore_Pd("Semaphore '{}' auto-signaled successfully".format(semaphore_key))
    except Exception as ex:
      self._semaphore_Pd("Error in auto-signal for '{}': {}".format(semaphore_key, ex))
      # Reset flag to allow retry on next iteration
      self._semaphore_signaled = False
    return


  def _setup_semaphore_env(self):
    """
    Hook method for plugins to set semaphore environment variables.

    Override this method in your plugin to expose environment variables
    to paired plugins (e.g., Container App Runners waiting for this plugin).

    This method is called automatically by _semaphore_maybe_auto_signal()
    right before signaling readiness.

    Example Usage
    -------------
    ```python
    def _setup_semaphore_env(self):
      # RedMesh example - expose API connection details
      host = getattr(self, 'cfg_host', None) or '127.0.0.1'
      port = self.cfg_port
      if port:
        self.semaphore_set_env('API_HOST', host)
        self.semaphore_set_env('API_PORT', str(port))
        self.semaphore_set_env('API_URL', 'http://{}:{}'.format(host, port))

      # Keysoft example - expose multiple endpoints
      localhost_ip = self.log.get_localhost_ip()
      port = getattr(self, 'cfg_port', 15033)
      self.semaphore_set_env('PORT', str(port))
      self.semaphore_set_env('RATIO1_AGENT_ENDPOINT',
        'http://{}:{}/query'.format(localhost_ip, port))

      # Container App Runner example - expose container details
      self.semaphore_set_env('CONTAINER_NAME', self._container_name)
      self.semaphore_set_env('CONTAINER_PORT', str(self._exposed_port))
    ```

    Returns
    -------
    None
    """
    # Default implementation is a no-op
    # Plugins override this method to set their specific env vars
    return


  def _semaphore_auto_cleanup(self):
    """
    Automatically clean up semaphore on plugin shutdown.

    This method is called automatically from the plugin's on_close() lifecycle.
    It clears the semaphore to signal to waiting plugins that this dependency
    is no longer available.

    Returns
    -------
    None
    """
    semaphore_key = getattr(self, 'cfg_semaphore', None)
    if not semaphore_key:
      return

    try:
      self.semaphore_clear()
      self._semaphore_Pd("Semaphore '{}' auto-cleanup completed".format(semaphore_key))
    except Exception as ex:
      self._semaphore_Pd("Error in semaphore auto-cleanup for '{}': {}".format(semaphore_key, ex))
    return


  def _semaphore_set_ready_flag(self):
    """
    Set the readiness flag to signal that this plugin is ready.

    This method delegates to set_plugin_ready(True) from _PluginReadinessMixin,
    which provides unified readiness for both semaphore and chainstore mechanisms.

    Use this method instead of directly setting readiness flags.
    This is typically used by plugins with custom readiness conditions
    (e.g., Container App Runner after health checks pass).

    The auto-signal mechanism will detect this flag and signal the semaphore
    automatically on the next process loop iteration.

    Returns
    -------
    None

    Example Usage
    -------------
    ```python
    # Container App Runner - after health check passes
    if probe_result:
      self.P("Health check passed - app is ready!", color='g')
      self._app_ready = True
      self._semaphore_set_ready_flag()  # Signal readiness for auto-signal
    ```
    """
    # Delegate to unified readiness (triggers both semaphore AND chainstore)
    self.set_plugin_ready(True)
    return


  def _semaphore_reset_signal(self):
    """
    Reset semaphore signaling state to allow re-signaling.

    This is useful for plugins that restart their services (e.g., Container App Runner
    restarting containers) and need to signal readiness again after restart.

    Calling this method will:
    - Clear the semaphore (signal to consumers that service is down)
    - Reset the _semaphore_signaled flag (allow re-signaling)
    - Reset unified readiness to False (deferred state)

    Returns
    -------
    None

    Example Usage
    -------------
    ```python
    # Container App Runner - when stopping container before restart
    def _stop_container_and_save_logs_to_disk(self):
      self.P(f"Stopping container app '{self.container_id}' ...")
      self._semaphore_reset_signal()  # Clear and reset for re-signaling
      # ... continue with container stop logic
    ```
    """
    # Clear the semaphore to signal consumers that service is down
    self.semaphore_clear()

    # Reset flags to allow re-signaling after restart
    self._semaphore_signaled = False

    # Reset unified readiness to deferred state (requires explicit set_plugin_ready(True))
    self.set_plugin_ready(False)

    return

  # ============================================================================
  # Consumer Methods (for CAR/WAR plugins that wait for dependencies)
  # ============================================================================

  def _semaphore_get_keys(self):
    """Get the list of semaphore keys this plugin waits for."""
    keys = getattr(self, 'cfg_semaphored_keys', None)
    return keys if keys else []

  def semaphore_is_ready(self, semaphore_key=None):
    """
    Check if a specific semaphore or all required semaphores are ready.

    Parameters
    ----------
    semaphore_key : str, optional
      Specific semaphore to check. If None, checks all SEMAPHORED_KEYS.

    Returns
    -------
    bool
      True if ready, False otherwise
    """
    if semaphore_key:
      # Check specific semaphore
      shmem_data = self.plugins_shmem.get(semaphore_key, {})
      return shmem_data.get('is_ready', False)

    # Check all required semaphores
    required_keys = self._semaphore_get_keys()
    if not required_keys:
      return True  # No dependencies, always ready

    for key in required_keys:
      shmem_data = self.plugins_shmem.get(key, {})
      if not shmem_data.get('is_ready', False):
        return False

    return True

  def semaphore_get_env(self):
    """
    Retrieve and aggregate environment variables from all semaphored keys.

    Returns
    -------
    dict
      Merged dictionary of all environment variables from ready semaphores
    """
    required_keys = self._semaphore_get_keys()
    if not required_keys:
      return {}

    result = {}
    for key in required_keys:
      shmem_data = self.plugins_shmem.get(key, {})
      if shmem_data.get('is_ready', False):
        env_vars = shmem_data.get('env', {})
        if env_vars:
          self._semaphore_Pd("Semaphore '{}' provides {} env vars: {}".format(
            key, len(env_vars), list(env_vars.keys())))
        result.update(env_vars)
      else:
        self._semaphore_Pd("Semaphore '{}' not ready, skipping env retrieval".format(key))

    return result


  def semaphore_get_env_value(self, semaphore_key, env_key, default=""):
    """
    Get a specific environment variable value from a specific semaphore.

    Parameters
    ----------
    semaphore_key : str
        The semaphore key to look up
    env_key : str
        The environment variable name
    default : str
        Default value if not found

    Returns
    -------
    str
        The value of the environment variable, or default if not found
    """
    shmem_data = self.plugins_shmem.get(semaphore_key, {})
    value = shmem_data.get('env', {}).get(env_key, default)
    return str(value) if value is not None else default


  def semaphore_get_env_value_by_path(self, full_key, default=""):
    """
    Get env value by full path string.

    Iterates through all configured semaphore keys and their env vars,
    constructing {semaphore_key}_{env_key} for each and comparing against
    the provided full_key.

    Parameters
    ----------
    full_key : str
        The complete key to look up (e.g. "MY_SEMAPHORE_CONTAINER_IP")
    default : str
        Default value if not found

    Returns
    -------
    str
        The env value, or default if not found
    """
    for semaphore_key in self._semaphore_get_keys():
      shmem_data = self.plugins_shmem.get(semaphore_key, {})
      if not shmem_data.get('is_ready', False):
        continue
      env_vars = shmem_data.get('env', {})
      for env_key, env_value in env_vars.items():
        constructed_key = self._semaphore_prefixed_key(semaphore_key, env_key)
        if constructed_key == full_key:
          return str(env_value)
    return default


  def semaphore_get_missing(self):
    """
    Get list of semaphores that are not yet ready.

    Returns
    -------
    list
      List of semaphore keys that are not ready
    """
    missing = []
    for key in self._semaphore_get_keys():
      if not self.semaphore_is_ready(key):
        missing.append(key)
    return missing

  def semaphore_get_status(self):
    """
    Get detailed status of all required semaphores.

    Returns
    -------
    dict
      Status information for each required semaphore
    """
    status = {}
    for key in self._semaphore_get_keys():
      shmem_data = self.plugins_shmem.get(key, {})
      if shmem_data:
        metadata = shmem_data.get('metadata', {})
        status[key] = {
          'ready': shmem_data.get('is_ready', False),
          'env_count': len(shmem_data.get('env', {})),
          'provider': metadata.get('plugin_signature'),
          'ready_since': metadata.get('ready_timestamp'),
        }
      else:
        status[key] = {
          'ready': False,
          'env_count': 0,
          'provider': None,
          'ready_since': None,
        }
    return status

  def semaphore_start_wait(self):
    """
    Mark the start of semaphore waiting period.

    Call this when beginning to wait for semaphores.
    """
    if self.__semaphore_wait_start is None:
      self.__semaphore_wait_start = tm()
      required_keys = self._semaphore_get_keys()
      self._semaphore_Pd("Starting wait for semaphores: {}".format(required_keys))
    return

  def semaphore_get_wait_elapsed(self):
    """
    Get elapsed time since waiting started.

    Returns
    -------
    float
      Elapsed time in seconds, or 0 if not waiting
    """
    if self.__semaphore_wait_start is None:
      return 0
    return tm() - self.__semaphore_wait_start

  def semaphore_reset_wait(self):
    """Reset the semaphore wait state (e.g., for retry after restart)."""
    self.__semaphore_wait_start = None
    self.__semaphore_ready_logged.clear()
    return

  def semaphore_check_with_logging(self):
    """
    Check semaphore status and log appropriately.

    Logs when individual semaphores become ready (only once per semaphore).

    Returns
    -------
    bool
      True if all semaphores are ready, False otherwise
    """
    required_keys = self._semaphore_get_keys()
    if not required_keys:
      return True

    all_ready = True
    for key in required_keys:
      is_ready = self.semaphore_is_ready(key)
      if is_ready and key not in self.__semaphore_ready_logged:
        shmem_data = self.plugins_shmem.get(key, {})
        metadata = shmem_data.get('metadata', {})
        provider = metadata.get('plugin_signature', 'unknown')
        env_count = len(shmem_data.get('env', {}))
        self._semaphore_Pd("Semaphore '{}' READY (provider: {}, env_vars: {})".format(
          key, provider, env_count))
        self.__semaphore_ready_logged.add(key)
      elif not is_ready:
        all_ready = False

    return all_ready



#!/usr/bin/env python3
"""
Passive Ratio1 SDK bandwidth capture for mainnet.

Primary metric: raw MQTT payload bytes as delivered to the SDK communicator.
Secondary metric: decoded heartbeat body composition after ENCODED_DATA
decompression, reported separately from raw bandwidth.
"""

from __future__ import annotations

import argparse
import hashlib
import json
import math
import os
from collections import Counter, defaultdict
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from threading import Lock
from time import sleep, time
from typing import Any

from ratio1.base.generic_session import GenericSession
from ratio1.comm.mqtt_wrapper import MQTTWrapper
from ratio1.const import comms as comm_ct
from ratio1.default.session.mqtt_session import MqttSession


LARGE_FIELD_THRESHOLD = 1024
MAX_TOP_ROWS = 20
DEFAULT_OUTPUT_DIR = "xperimental/payloads_tests/evidence/raw_bandwidth"


def utc_now_iso() -> str:
  return datetime.now(timezone.utc).isoformat(timespec="seconds")


def compact_json_text(value: Any) -> str:
  return json.dumps(value, ensure_ascii=True, separators=(",", ":"))


def compact_json_bytes(value: Any) -> int:
  return len(compact_json_text(value).encode("utf-8"))


def safe_text_size(value: str | None) -> int:
  if value is None:
    return 0
  return len(value.encode("utf-8"))


def human_bytes(num_bytes: float) -> str:
  value = float(num_bytes)
  for unit in ["B", "KB", "MB", "GB", "TB"]:
    if value < 1024.0 or unit == "TB":
      if unit == "B":
        return f"{int(value)}{unit}"
      return f"{value:.1f}{unit}"
    value /= 1024.0
  return f"{value:.1f}TB"


def pct(part: float, whole: float) -> float:
  if whole <= 0:
    return 0.0
  return round(100.0 * part / whole, 1)


def percentile(values: list[int], q: float) -> int:
  if not values:
    return 0
  if len(values) == 1:
    return values[0]
  ordered = sorted(values)
  idx = int(math.ceil((q / 100.0) * len(ordered))) - 1
  idx = max(0, min(idx, len(ordered) - 1))
  return ordered[idx]


def compact_preview(value: Any, max_str: int = 80) -> str:
  if isinstance(value, dict):
    return f"<dict {len(value)} keys, {compact_json_bytes(value)}B>"
  if isinstance(value, list):
    return f"<list {len(value)} items, {compact_json_bytes(value)}B>"
  if isinstance(value, str):
    trimmed = value[:max_str]
    if len(value) > max_str:
      trimmed += "..."
    return f"<string {len(value.encode('utf-8'))}B> {trimmed}"
  return repr(value)


def normalize_path(value: Any) -> list[Any]:
  if isinstance(value, list):
    result = list(value[:4])
    while len(result) < 4:
      result.append(None)
    return result
  return [None, None, None, None]


def render_sender_label(sender_id: str | None, sender_addr: str | None) -> str | None:
  if sender_id and sender_addr:
    return f"{sender_id}<{sender_addr}>"
  return sender_id or sender_addr


def message_class(event_type: str | None, signature: str | None, raw_obj: dict[str, Any]) -> str:
  ev = (event_type or "").upper()
  if ev == "HEARTBEAT":
    return "heartbeat"
  if ev == "NOTIFICATION":
    notif_type = raw_obj.get("NOTIFICATION_TYPE") or raw_obj.get("STATUS_TYPE") or "-"
    return f"notification:{notif_type}"
  if ev == "PAYLOAD":
    return f"payload:{signature or '-'}"
  if signature:
    return f"payload:{signature}"
  return ev.lower() or "unknown"


def count_prefixed_fields(obj: dict[str, Any], prefix: str) -> int:
  return sum(1 for key in obj if isinstance(key, str) and key.startswith(prefix))


def collect_presence(obj: dict[str, Any]) -> dict[str, bool]:
  keys = set(obj.keys())
  return {
    "IMG": "IMG" in keys,
    "IMG_ORIG": "IMG_ORIG" in keys,
    "HISTORY": "HISTORY" in keys,
    "DCT_STATS": "DCT_STATS" in keys,
    "COMM_STATS": "COMM_STATS" in keys,
    "ACTIVE_PLUGINS": "ACTIVE_PLUGINS" in keys,
    "CONFIG_STREAMS": "CONFIG_STREAMS" in keys,
    "EE_WHITELIST": "EE_WHITELIST" in keys,
    "TAGS": "TAGS" in keys,
    "ID_TAGS": "ID_TAGS" in keys,
  }


def default_like(value: Any) -> bool:
  return value in (None, "", [], {}, False)


def summarize_raw_object(raw_obj: dict[str, Any], raw_bytes: int, recv_ts: float, comm_type: str, topic: str, log) -> dict[str, Any]:
  path = normalize_path(raw_obj.get("EE_PAYLOAD_PATH"))
  node, stream, signature, instance = path
  event_type = raw_obj.get("EE_EVENT_TYPE")
  sender_addr = raw_obj.get("EE_SENDER")
  sender_id = raw_obj.get("EE_ID")
  field_sizes = {key: compact_json_bytes(value) for key, value in raw_obj.items()}
  large_fields = sorted(
    [{"field": key, "bytes": size} for key, size in field_sizes.items() if size >= LARGE_FIELD_THRESHOLD],
    key=lambda item: item["bytes"],
    reverse=True,
  )
  preview_fields = {}
  for item in large_fields[:5]:
    preview_fields[item["field"]] = compact_preview(raw_obj.get(item["field"]))

  shape_key = "|".join(sorted(str(key) for key in raw_obj.keys()))
  row = {
    "recv_ts": recv_ts,
    "recv_iso": datetime.fromtimestamp(recv_ts, tz=timezone.utc).isoformat(timespec="seconds"),
    "comm_type": comm_type,
    "topic": topic,
    "raw_bytes": raw_bytes,
    "sender": render_sender_label(sender_id, sender_addr),
    "sender_addr": sender_addr,
    "sender_id": sender_id,
    "destination": raw_obj.get("EE_DESTINATION") or raw_obj.get("EE_DEST"),
    "event_type": event_type,
    "stream": stream,
    "signature": signature,
    "instance": instance,
    "message_class": message_class(event_type, signature, raw_obj),
    "encrypted": bool(raw_obj.get("EE_IS_ENCRYPTED", False)),
    "top_level_keys": sorted(str(key) for key in raw_obj.keys()),
    "top_level_key_count": len(raw_obj),
    "raw_field_sizes": field_sizes,
    "large_raw_fields": large_fields,
    "raw_preview": preview_fields,
    "_c_field_count": count_prefixed_fields(raw_obj, "_C_"),
    "_p_field_count": count_prefixed_fields(raw_obj, "_P_"),
    "presence": collect_presence(raw_obj),
    "empty_or_default_fields": sorted(key for key, value in raw_obj.items() if default_like(value)),
    "shape_hash": hashlib.sha1(shape_key.encode("utf-8")).hexdigest()[:10],
  }

  if row["message_class"] == "heartbeat":
    hb_version = raw_obj.get("HEARTBEAT_VERSION")
    hb_version_norm = str(hb_version).lower() if hb_version is not None else None
    encoded_data = raw_obj.get("ENCODED_DATA")
    row["heartbeat_version"] = hb_version
    row["heartbeat_encoded_text_bytes"] = safe_text_size(encoded_data) if isinstance(encoded_data, str) else 0
    row["heartbeat_decoded_bytes"] = 0
    row["heartbeat_inner_field_sizes"] = {}
    row["heartbeat_inner_large_fields"] = []
    if hb_version_norm == "v2" and isinstance(encoded_data, str):
      try:
        hb_inner = json.loads(log.decompress_text(encoded_data))
        inner_field_sizes = {key: compact_json_bytes(value) for key, value in hb_inner.items()}
        row["heartbeat_decoded_bytes"] = compact_json_bytes(hb_inner)
        row["heartbeat_inner_field_sizes"] = inner_field_sizes
        row["heartbeat_inner_large_fields"] = sorted(
          [{"field": key, "bytes": size} for key, size in inner_field_sizes.items() if size >= LARGE_FIELD_THRESHOLD],
          key=lambda item: item["bytes"],
          reverse=True,
        )
      except Exception as exc:
        row["heartbeat_decode_error"] = str(exc)

  return row


@dataclass
class CaptureArtifacts:
  capture_file: Path
  summary_json: Path
  results_md: Path
  metadata_json: Path


class JsonlWriter:
  def __init__(self, path: Path):
    self.path = path
    self.path.parent.mkdir(parents=True, exist_ok=True)
    self._handle = self.path.open("w", encoding="utf-8")
    self._lock = Lock()

  def write(self, row: dict[str, Any]) -> None:
    with self._lock:
      self._handle.write(json.dumps(row, ensure_ascii=True) + "\n")
      self._handle.flush()

  def close(self) -> None:
    with self._lock:
      self._handle.close()


class RecordingMQTTWrapper(MQTTWrapper):
  def __init__(self, *args, record_callback=None, **kwargs):
    self._record_callback = record_callback
    super().__init__(*args, **kwargs)

  def _callback_on_message(self, client, userdata, message, *args, **kwargs):
    if self._record_callback is not None:
      try:
        self._record_callback(
          comm_type=self._comm_type,
          topic=message.topic,
          payload_bytes=bytes(message.payload),
          log=self.log,
        )
      except Exception:
        pass
    return super()._callback_on_message(client, userdata, message, *args, **kwargs)


class PassiveBandwidthMqttSession(MqttSession):
  def __init__(self, *args, raw_record_callback=None, **kwargs):
    self._raw_record_callback = raw_record_callback
    super().__init__(*args, **kwargs)

  def _make_wrapper(self, *, send_channel_name=None, recv_channel_name=None, comm_type=None, recv_buff=None):
    return RecordingMQTTWrapper(
      log=self.log,
      config=self._config,
      send_channel_name=send_channel_name,
      recv_channel_name=recv_channel_name,
      comm_type=comm_type,
      recv_buff=recv_buff,
      connection_name=self.name,
      verbosity=self._verbosity,
      record_callback=self._raw_record_callback,
    )

  def startup(self):
    self._default_communicator = self._make_wrapper(
      send_channel_name=comm_ct.COMMUNICATION_PAYLOADS_CHANNEL,
      recv_channel_name=comm_ct.COMMUNICATION_PAYLOADS_CHANNEL,
      comm_type=comm_ct.COMMUNICATION_DEFAULT,
      recv_buff=self._payload_messages,
    )
    self._heartbeats_communicator = self._make_wrapper(
      send_channel_name=comm_ct.COMMUNICATION_CONFIG_CHANNEL,
      recv_channel_name=comm_ct.COMMUNICATION_CTRL_CHANNEL,
      comm_type=comm_ct.COMMUNICATION_HEARTBEATS,
      recv_buff=self._hb_messages,
    )
    self._notifications_communicator = self._make_wrapper(
      recv_channel_name=comm_ct.COMMUNICATION_NOTIF_CHANNEL,
      comm_type=comm_ct.COMMUNICATION_NOTIFICATIONS,
      recv_buff=self._notif_messages,
    )
    self._MqttSession__communicators = {
      "default": self._default_communicator,
      "heartbeats": self._heartbeats_communicator,
      "notifications": self._notifications_communicator,
    }
    return GenericSession.startup(self)

  def _GenericSession__request_pipelines_from_net_config_monitor(self, node_addr=None):
    if not hasattr(self, "_passive_netconfig_warning_shown"):
      self._passive_netconfig_warning_shown = True
      self.P("Passive bandwidth capture: disabling automatic net-config requests.", color="y")
    return


class CaptureRecorder:
  def __init__(self, jsonl_path: Path):
    self.writer = JsonlWriter(jsonl_path)
    self.lock = Lock()
    self.count = 0
    self.first_recv_ts = None
    self.last_recv_ts = None

  def record(self, *, comm_type: str, topic: str, payload_bytes: bytes, log) -> None:
    recv_ts = time()
    raw_bytes = len(payload_bytes)
    try:
      raw_text = payload_bytes.decode("utf-8")
      raw_obj = json.loads(raw_text)
    except Exception as exc:
      row = {
        "recv_ts": recv_ts,
        "recv_iso": datetime.fromtimestamp(recv_ts, tz=timezone.utc).isoformat(timespec="seconds"),
        "comm_type": comm_type,
        "topic": topic,
        "raw_bytes": raw_bytes,
        "parse_error": str(exc),
      }
    else:
      row = summarize_raw_object(
        raw_obj=raw_obj,
        raw_bytes=raw_bytes,
        recv_ts=recv_ts,
        comm_type=comm_type,
        topic=topic,
        log=log,
      )
    self.writer.write(row)
    with self.lock:
      self.count += 1
      if self.first_recv_ts is None:
        self.first_recv_ts = recv_ts
      self.last_recv_ts = recv_ts

  def close(self) -> None:
    self.writer.close()


def top_n_rows(counter_like: dict[str, dict[str, Any]], sort_key: str, limit: int = MAX_TOP_ROWS) -> list[dict[str, Any]]:
  rows = list(counter_like.values())
  rows.sort(key=lambda item: item.get(sort_key, 0), reverse=True)
  return rows[:limit]


def update_field_totals(total_sizes: dict[str, int], total_counts: dict[str, int], field_sizes: dict[str, int]) -> None:
  for field, size in field_sizes.items():
    total_sizes[field] += size
    total_counts[field] += 1


def render_table(headers: list[str], rows: list[list[Any]]) -> str:
  lines = [
    "| " + " | ".join(headers) + " |",
    "| " + " | ".join(["---"] * len(headers)) + " |",
  ]
  for row in rows:
    lines.append("| " + " | ".join(str(cell) for cell in row) + " |")
  return "\n".join(lines)


def analyze_capture(capture_file: Path, metadata_path: Path, results_md: Path, summary_json: Path, args: argparse.Namespace) -> dict[str, Any]:
  rows = []
  with capture_file.open("r", encoding="utf-8") as handle:
    for line in handle:
      rows.append(json.loads(line))

  metadata = json.loads(metadata_path.read_text(encoding="utf-8"))
  duration = float(metadata["duration_seconds"])
  if duration <= 0 and rows:
    duration = max(0.001, rows[-1]["recv_ts"] - rows[0]["recv_ts"])
  total_raw_bytes = sum(row.get("raw_bytes", 0) for row in rows)
  total_messages = len(rows)
  bytes_per_second = total_raw_bytes / duration if duration > 0 else 0.0

  by_class = {}
  by_sender = {}
  by_stream_sig = {}
  raw_field_sizes = defaultdict(int)
  raw_field_counts = defaultdict(int)
  hb_inner_field_sizes = defaultdict(int)
  hb_inner_field_counts = defaultdict(int)
  empty_field_sizes = defaultdict(int)
  shape_counter = Counter()
  first_half_bytes = 0
  second_half_bytes = 0
  first_half_msgs = 0
  second_half_msgs = 0
  minute_bins = defaultdict(lambda: defaultdict(int))
  heartbeat_decoded_total = 0

  if rows:
    start_ts = rows[0]["recv_ts"]
    end_ts = rows[-1]["recv_ts"]
  else:
    start_ts = metadata["start_ts"]
    end_ts = metadata["end_ts"]
  midpoint = start_ts + ((end_ts - start_ts) / 2.0 if end_ts >= start_ts else 0.0)

  for row in rows:
    raw_bytes = row.get("raw_bytes", 0)
    msg_class = row.get("message_class", "unknown")
    sender = row.get("sender") or "unknown"
    stream = row.get("stream") or "-"
    signature = row.get("signature") or "-"
    recv_ts = row.get("recv_ts", start_ts)

    by_class.setdefault(msg_class, {"message_class": msg_class, "count": 0, "total_bytes": 0, "sizes": []})
    by_class[msg_class]["count"] += 1
    by_class[msg_class]["total_bytes"] += raw_bytes
    by_class[msg_class]["sizes"].append(raw_bytes)

    by_sender.setdefault(sender, {"sender": sender, "count": 0, "total_bytes": 0})
    by_sender[sender]["count"] += 1
    by_sender[sender]["total_bytes"] += raw_bytes

    stream_sig_key = f"{stream} / {signature}"
    by_stream_sig.setdefault(stream_sig_key, {"stream_signature": stream_sig_key, "count": 0, "total_bytes": 0})
    by_stream_sig[stream_sig_key]["count"] += 1
    by_stream_sig[stream_sig_key]["total_bytes"] += raw_bytes

    update_field_totals(raw_field_sizes, raw_field_counts, row.get("raw_field_sizes", {}))
    update_field_totals(hb_inner_field_sizes, hb_inner_field_counts, row.get("heartbeat_inner_field_sizes", {}))

    for field in row.get("empty_or_default_fields", []):
      size = row.get("raw_field_sizes", {}).get(field, 0)
      empty_field_sizes[field] += size

    heartbeat_decoded_total += row.get("heartbeat_decoded_bytes", 0)
    shape_counter[row.get("shape_hash", "missing")] += 1

    minute_idx = int((recv_ts - start_ts) // 60) if rows else 0
    minute_bins[minute_idx][msg_class] += raw_bytes

    if recv_ts <= midpoint:
      first_half_msgs += 1
      first_half_bytes += raw_bytes
    else:
      second_half_msgs += 1
      second_half_bytes += raw_bytes

  by_class_rows = []
  for item in by_class.values():
    sizes = item.pop("sizes")
    item["avg_bytes"] = round(item["total_bytes"] / item["count"], 1) if item["count"] else 0
    item["p95_bytes"] = percentile(sizes, 95)
    item["max_bytes"] = max(sizes) if sizes else 0
    by_class_rows.append(item)
  by_class_rows.sort(key=lambda item: item["total_bytes"], reverse=True)

  top_raw_fields = sorted(
    [
      {
        "field": field,
        "messages_present": raw_field_counts[field],
        "total_bytes": size,
        "avg_when_present": round(size / raw_field_counts[field], 1) if raw_field_counts[field] else 0,
      }
      for field, size in raw_field_sizes.items()
    ],
    key=lambda item: item["total_bytes"],
    reverse=True,
  )

  top_hb_inner_fields = sorted(
    [
      {
        "field": field,
        "messages_present": hb_inner_field_counts[field],
        "total_bytes": size,
        "avg_when_present": round(size / hb_inner_field_counts[field], 1) if hb_inner_field_counts[field] else 0,
      }
      for field, size in hb_inner_field_sizes.items()
    ],
    key=lambda item: item["total_bytes"],
    reverse=True,
  )

  top_empty_fields = sorted(
    [{"field": field, "total_bytes": size} for field, size in empty_field_sizes.items()],
    key=lambda item: item["total_bytes"],
    reverse=True,
  )

  largest_rows = sorted(rows, key=lambda row: row.get("raw_bytes", 0), reverse=True)[:5]
  top_shapes = shape_counter.most_common(5)

  results = {
    "scope": {
      "script_path": "xperimental/payloads_tests/sdk_bandwidth_capture.py",
      "capture_file": str(capture_file),
      "summary_json": str(summary_json),
      "results_md": str(results_md),
      "network": args.network,
      "duration_seconds": duration,
      "message_count": total_messages,
      "stop_reason": metadata["stop_reason"],
    },
    "totals": {
      "raw_bytes": total_raw_bytes,
      "bytes_per_second": bytes_per_second,
      "bytes_per_minute": bytes_per_second * 60.0,
      "heartbeat_decoded_total": heartbeat_decoded_total,
      "first_half_messages": first_half_msgs,
      "first_half_bytes": first_half_bytes,
      "second_half_messages": second_half_msgs,
      "second_half_bytes": second_half_bytes,
    },
    "by_class": by_class_rows,
    "by_sender": top_n_rows(by_sender, "total_bytes"),
    "by_stream_signature": top_n_rows(by_stream_sig, "total_bytes"),
    "top_raw_fields": top_raw_fields[:MAX_TOP_ROWS],
    "top_heartbeat_inner_fields": top_hb_inner_fields[:MAX_TOP_ROWS],
    "top_empty_fields": top_empty_fields[:MAX_TOP_ROWS],
    "largest_rows": largest_rows,
    "top_shapes": top_shapes,
    "minute_bins": {str(idx): values for idx, values in sorted(minute_bins.items())},
  }

  summary_json.write_text(json.dumps(results, indent=2, ensure_ascii=True), encoding="utf-8")

  msg_class_table = render_table(
    ["message class", "count", "total raw bytes", "avg bytes", "p95 bytes", "max bytes", "byte share"],
    [
      [
        item["message_class"],
        item["count"],
        human_bytes(item["total_bytes"]),
        int(item["avg_bytes"]),
        item["p95_bytes"],
        item["max_bytes"],
        f"{pct(item['total_bytes'], total_raw_bytes)}%",
      ]
      for item in by_class_rows[:MAX_TOP_ROWS]
    ],
  )

  sender_table = render_table(
    ["sender", "count", "total raw bytes", "avg bytes"],
    [
      [
        item["sender"],
        item["count"],
        human_bytes(item["total_bytes"]),
        int(item["total_bytes"] / item["count"]) if item["count"] else 0,
      ]
      for item in top_n_rows(by_sender, "total_bytes")[:10]
    ],
  )

  stream_table = render_table(
    ["stream / signature", "count", "total raw bytes", "avg bytes"],
    [
      [
        item["stream_signature"],
        item["count"],
        human_bytes(item["total_bytes"]),
        int(item["total_bytes"] / item["count"]) if item["count"] else 0,
      ]
      for item in top_n_rows(by_stream_sig, "total_bytes")[:10]
    ],
  )

  raw_field_table = render_table(
    ["field", "messages present", "total estimated raw bytes", "avg bytes when present"],
    [
      [
        item["field"],
        item["messages_present"],
        human_bytes(item["total_bytes"]),
        int(item["avg_when_present"]),
      ]
      for item in top_raw_fields[:15]
    ],
  )

  hb_inner_field_table = render_table(
    ["field", "messages present", "total decoded bytes", "avg bytes when present"],
    [
      [
        item["field"],
        item["messages_present"],
        human_bytes(item["total_bytes"]),
        int(item["avg_when_present"]),
      ]
      for item in top_hb_inner_fields[:15]
    ],
  )

  largest_examples = []
  for idx, row in enumerate(largest_rows, start=1):
    largest_examples.append(
      f"- Example {idx}: `{row.get('message_class')}` from `{row.get('sender')}` "
      f"({human_bytes(row.get('raw_bytes', 0))}, stream=`{row.get('stream')}`, signature=`{row.get('signature')}`)\n"
      f"  large raw fields: " +
      ", ".join(f"{item['field']}={human_bytes(item['bytes'])}" for item in row.get("large_raw_fields", [])[:5]) + "\n"
      f"  preview: `{json.dumps(row.get('raw_preview', {}), ensure_ascii=True)}`"
    )

  minute_lines = []
  for minute_idx, values in sorted(minute_bins.items()):
    top_class = sorted(values.items(), key=lambda item: item[1], reverse=True)[:3]
    rendered = ", ".join(f"{name}={human_bytes(size)}" for name, size in top_class)
    minute_lines.append(f"- minute {minute_idx}: {rendered}")

  md = f"""# Ratio1 SDK Raw Bandwidth Capture Results

## Scope
- script path: `xperimental/payloads_tests/sdk_bandwidth_capture.py`
- capture command: `python3 xperimental/payloads_tests/sdk_bandwidth_capture.py --seconds {args.seconds} --max-messages {args.max_messages}`
- analysis command: `python3 xperimental/payloads_tests/sdk_bandwidth_capture.py --analyze-only --capture-file {capture_file}`
- network: `{args.network}`
- capture window: `{metadata['start_iso']}` to `{metadata['end_iso']}` ({duration:.1f}s)
- message count: `{total_messages}`
- stop reason: `{metadata['stop_reason']}`

## Measurement Rules
- Primary bandwidth metric is raw MQTT payload size: `len(message.payload)` captured before SDK parsing or heartbeat decompression.
- Heartbeat decoded-body analysis is reported separately and is **not** counted as raw bandwidth.
- MQTT topic/header overhead is still excluded because the SDK callback surface does not expose it.
- The session disables automatic SDK net-config requests to keep the run passive.

## Executive Summary
- Raw MQTT payload throughput averaged `{human_bytes(bytes_per_second)}/s` (`{human_bytes(bytes_per_second * 60.0)}/min`).
- Raw message bandwidth was dominated by `{by_class_rows[0]['message_class'] if by_class_rows else 'n/a'}` at `{pct(by_class_rows[0]['total_bytes'], total_raw_bytes) if by_class_rows else 0}%` of observed bytes.
- Heartbeat decoded bodies expanded to `{human_bytes(heartbeat_decoded_total)}` after `ENCODED_DATA` decompression, but those bytes were **not** used in the raw bandwidth totals.
- First half vs second half raw bytes: `{human_bytes(first_half_bytes)}` vs `{human_bytes(second_half_bytes)}`.

## Raw Byte Distribution
{msg_class_table}

{sender_table}

{stream_table}

## Top Raw Fields
{raw_field_table}

## Heartbeat Decoded Composition
These fields describe what sits inside compressed heartbeat bodies. They are useful for optimization analysis but are not additive with the raw heartbeat bytes above.

{hb_inner_field_table}

## Empty or Default-like Raw Fields
{render_table(
  ["field", "total estimated raw bytes"],
  [[item["field"], human_bytes(item["total_bytes"])] for item in top_empty_fields[:10]]
)}

## Largest Raw Message Examples
{os.linesep.join(largest_examples)}

## Stability Check
- first half message count / raw bytes: `{first_half_msgs}` / `{human_bytes(first_half_bytes)}`
- second half message count / raw bytes: `{second_half_msgs}` / `{human_bytes(second_half_bytes)}`
- top shape hashes: `{", ".join(f"{shape}:{count}" for shape, count in top_shapes)}`

## Raw Bytes By Minute
{os.linesep.join(minute_lines)}

## Artifacts
- capture jsonl: `{capture_file}`
- summary json: `{summary_json}`
- results md: `{results_md}`

## Verification
- command: `python3 xperimental/payloads_tests/sdk_bandwidth_capture.py --seconds {args.seconds} --max-messages {args.max_messages}`
  result: `pass`
  evidence: `Captured {total_messages} raw MQTT payloads over {duration:.1f}s on {args.network}`
"""
  results_md.write_text(md, encoding="utf-8")
  return results


def build_artifacts(output_dir: Path) -> CaptureArtifacts:
  ts = datetime.now(timezone.utc).strftime("%Y%m%dT%H%M%S%z")
  output_dir.mkdir(parents=True, exist_ok=True)
  base = output_dir / f"{ts}_mainnet_bandwidth"
  return CaptureArtifacts(
    capture_file=base.with_suffix(".jsonl"),
    summary_json=Path(str(base) + "_summary.json"),
    results_md=Path(str(base) + "_results.md"),
    metadata_json=Path(str(base) + "_metadata.json"),
  )


def parse_args() -> argparse.Namespace:
  parser = argparse.ArgumentParser(description="Passive Ratio1 SDK raw-bandwidth capture.")
  parser.add_argument("--seconds", type=int, default=600, help="Capture duration in seconds.")
  parser.add_argument("--max-messages", type=int, default=30000, help="Maximum number of messages to capture.")
  parser.add_argument("--network", default="mainnet", help="EVM network for the SDK session.")
  parser.add_argument("--output-dir", default=DEFAULT_OUTPUT_DIR, help="Directory for local artifacts.")
  parser.add_argument("--session-name", default="sdk-bandwidth-capture", help="SDK session alias.")
  parser.add_argument("--analyze-only", action="store_true", help="Skip live capture and analyze an existing JSONL file.")
  parser.add_argument("--capture-file", help="Existing capture JSONL to analyze.")
  parser.add_argument("--metadata-file", help="Existing capture metadata JSON to analyze.")
  parser.add_argument("--results-md", help="Explicit markdown output path.")
  parser.add_argument("--summary-json", help="Explicit summary json output path.")
  parser.add_argument("--silent", action="store_true", help="Reduce SDK logs during live capture.")
  return parser.parse_args()


def run_capture(args: argparse.Namespace) -> CaptureArtifacts:
  artifacts = build_artifacts(Path(args.output_dir))
  recorder = CaptureRecorder(artifacts.capture_file)
  start_ts = time()
  metadata = {
    "start_ts": start_ts,
    "start_iso": utc_now_iso(),
    "network": args.network,
    "session_name": args.session_name,
    "stop_reason": "running",
  }
  artifacts.metadata_json.write_text(json.dumps(metadata, indent=2), encoding="utf-8")

  session = None
  try:
    session = PassiveBandwidthMqttSession(
      name=args.session_name,
      silent=args.silent,
      auto_configuration=True,
      run_dauth=True,
      use_home_folder=False,
      local_cache_base_folder=".",
      evm_network=args.network,
      raw_record_callback=recorder.record,
    )
    deadline = start_ts + args.seconds
    while time() < deadline and recorder.count < args.max_messages:
      sleep(0.25)
    metadata["stop_reason"] = "max-messages" if recorder.count >= args.max_messages else "time-window"
  finally:
    metadata["end_ts"] = time()
    metadata["end_iso"] = utc_now_iso()
    metadata["duration_seconds"] = round(metadata["end_ts"] - start_ts, 3)
    metadata["message_count"] = recorder.count
    artifacts.metadata_json.write_text(json.dumps(metadata, indent=2), encoding="utf-8")
    recorder.close()
    if session is not None:
      session.close(wait_close=True)
  return artifacts


def main() -> int:
  args = parse_args()
  if args.analyze_only:
    if not args.capture_file:
      raise SystemExit("--capture-file is required with --analyze-only")
    capture_file = Path(args.capture_file)
    metadata_file = Path(args.metadata_file) if args.metadata_file else capture_file.with_name(capture_file.stem.replace("_results", "").replace("_summary", "") + "_metadata.json")
    results_md = Path(args.results_md) if args.results_md else capture_file.with_name(capture_file.stem + "_results.md")
    summary_json = Path(args.summary_json) if args.summary_json else capture_file.with_name(capture_file.stem + "_summary.json")
    analyze_capture(capture_file, metadata_file, results_md, summary_json, args)
    print(results_md)
    return 0

  artifacts = run_capture(args)
  analyze_capture(artifacts.capture_file, artifacts.metadata_json, artifacts.results_md, artifacts.summary_json, args)
  print(str(artifacts.results_md))
  return 0


if __name__ == "__main__":
  raise SystemExit(main())
